from ..types.clob_types import ContractConfig

CONFIG = {
    137: ContractConfig(
        exchange="0x4bFb41d5B3570DeFd03C39a9A4D8dE6Bd8B8982E",
        collateral="0x2791Bca1f2de4661ED88A30C99A7a9449Aa84174",
        conditional_tokens="0x4D97DCd97eC945f40cF65F87097ACe5EA0476045",
    ),
    80002: ContractConfig(
        exchange="0xdFE02Eb6733538f8Ea35D585af8DE5958AD99E40",
        collateral="0x9c4e1703476e875070ee25b56a58b008cfb8fa78",
        conditional_tokens="0x69308FB512518e39F9b16112fA8d994F4e2Bf8bB",
    ),
}

NEG_RISK_CONFIG = {
    137: ContractConfig(
        exchange="0xC5d563A36AE78145C45a50134d48A1215220f80a",
        collateral="0x2791bca1f2de4661ed88a30c99a7a9449aa84174",
        conditional_tokens="0x4D97DCd97eC945f40cF65F87097ACe5EA0476045",
    ),
    80002: ContractConfig(
        exchange="0xd91E80cF2E7be2e162c6513ceD06f1dD0dA35296",
        collateral="0x9c4e1703476e875070ee25b56a58b008cfb8fa78",
        conditional_tokens="0x69308FB512518e39F9b16112fA8d994F4e2Bf8bB",
    ),
}

def get_contract_config(chain_id: int, neg_risk: bool = False) -> ContractConfig:
    """Get the contract configuration for the chain."""
    config = NEG_RISK_CONFIG.get(chain_id) if neg_risk else CONFIG.get(chain_id)
    if config is None:
        msg = f"Invalid chain_id: ${chain_id}"
        raise ValueError(msg)

    return config


GRAPHQL_ENDPOINTS = {
    "activity_subgraph": "https://api.goldsky.com/api/public/project_cl6mb8i9h0003e201j6li0diw/subgraphs/activity-subgraph/0.0.4/gn",
    "fpmm_subgraph": "https://api.goldsky.com/api/public/project_cl6mb8i9h0003e201j6li0diw/subgraphs/fpmm-subgraph/0.0.1/gn",
    "open_interest_subgraph": "https://api.goldsky.com/api/public/project_cl6mb8i9h0003e201j6li0diw/subgraphs/oi-subgraph/0.0.6/gn",
    "orderbook_subgraph": "https://api.goldsky.com/api/public/project_cl6mb8i9h0003e201j6li0diw/subgraphs/orderbook-subgraph/0.0.1/gn",
    "pnl_subgraph": "https://api.goldsky.com/api/public/project_cl6mb8i9h0003e201j6li0diw/subgraphs/pnl-subgraph/0.0.14/gn",
    "positions_subgraph": "https://api.goldsky.com/api/public/project_cl6mb8i9h0003e201j6li0diw/subgraphs/positions-subgraph/0.0.7/gn",
    "sports_oracle_subgraph": "https://api.goldsky.com/api/public/project_cl6mb8i9h0003e201j6li0diw/subgraphs/sports-oracle-subgraph/0.0.1/gn",
    "wallet_subgraph": "https://api.goldsky.com/api/public/project_cl6mb8i9h0003e201j6li0diw/subgraphs/wallet-subgraph/0.0.4/gn",
}



from json import load
from pathlib import Path
from typing import Literal, Optional

from web3 import Web3
from web3.exceptions import ContractCustomError
from web3.middleware import ExtraDataToPOAMiddleware, SignAndSendRawMiddlewareBuilder

from ..types.common import EthAddress, Keccak256
from ..utilities.config import get_contract_config
from ..utilities.constants import HASH_ZERO, POLYGON
from ..utilities.web3.abis.custom_contract_errors import CUSTOM_ERROR_DICT
from ..utilities.web3.helpers import get_index_set


def _load_abi(contract_name: str) -> list:
    abi_path = Path(__file__).parent.parent/"utilities"/"web3"/"abis"/f"{contract_name}.json"
    with Path.open(abi_path) as f:
        return load(f)

class PolymarketWeb3Client:
    def __init__(self, private_key: str , chain_id: Literal[137, 80002] = POLYGON):

        self.w3 = Web3(Web3.HTTPProvider("https://polygon-rpc.com"))
        self.w3.middleware_onion.inject(ExtraDataToPOAMiddleware, layer=0)
        self.w3.middleware_onion.inject(SignAndSendRawMiddlewareBuilder.build(private_key), layer=0)

        self.account = self.w3.eth.account.from_key(private_key)

        self.config = get_contract_config(chain_id, neg_risk=False)
        self.neg_risk_config = get_contract_config(chain_id, neg_risk=True)

        self.usdc_address = Web3.to_checksum_address(self.config.collateral)
        self.usdc_abi = _load_abi("UChildERC20Proxy")
        self.usdc = self.contract(self.usdc_address, self.usdc_abi)

        self.conditional_tokens_address = Web3.to_checksum_address(self.config.conditional_tokens)
        self.conditional_tokens_abi = _load_abi("ConditionalTokens")
        self.conditional_tokens = self.contract(self.conditional_tokens_address, self.conditional_tokens_abi)

        self.exchange_address = Web3.to_checksum_address(self.config.exchange)
        self.exchange_abi = _load_abi("CTFExchange")
        self.exchange = self.contract(self.exchange_address, self.exchange_abi)

        self.neg_risk_exchange_address = Web3.to_checksum_address(self.neg_risk_config.exchange)
        self.neg_risk_exchange_abi = _load_abi("NegRiskCtfExchange")
        self.neg_risk_exchange = self.contract(self.neg_risk_exchange_address, self.neg_risk_exchange_abi)

        self.neg_risk_adapter_address = "0xd91E80cF2E7be2e162c6513ceD06f1dD0dA35296"
        self.neg_risk_adapter_abi = _load_abi("NegRiskAdapter")
        self.neg_risk_adapter = self.contract(self.neg_risk_adapter_address, self.neg_risk_adapter_abi)

        self.proxy_factory_address = "0xaB45c5A4B0c941a2F231C04C3f49182e1A254052"
        self.proxy_factory_abi = _load_abi("ProxyWalletFactory")
        self.proxy_factory = self.contract(self.proxy_factory_address, self.proxy_factory_abi)

    def _encode_split(self, condition_id: Keccak256, amount: int) -> str:
        return self.conditional_tokens.encode_abi(
            abi_element_identifier="splitPosition",
            args=[self.usdc_address, HASH_ZERO, condition_id, [1, 2], amount],
        )

    def _encode_merge(self, condition_id: Keccak256, amount: int) -> str:
        return self.conditional_tokens.encode_abi(
            abi_element_identifier="mergePositions",
            args=[self.usdc_address, HASH_ZERO, condition_id, [1, 2], amount],
        )

    def _encode_redeem(self, condition_id: Keccak256) -> str:
        return self.conditional_tokens.encode_abi(
            abi_element_identifier="redeemPositions",
            args=[self.usdc_address, HASH_ZERO, condition_id, [1, 2]],
        )

    def _encode_redeem_neg_risk(self, condition_id: Keccak256, amounts: list[int]) -> str:
        return self.neg_risk_adapter.encode_abi(
            abi_element_identifier="redeemPositions",
            args=[condition_id, amounts],
        )
    def _encode_convert(self, neg_risk_market_id: Keccak256, index_set: int, amount: int) -> str:
        return self.neg_risk_adapter.encode_abi(
            abi_element_identifier="convertPositions",
            args=[neg_risk_market_id, index_set, amount],
        )

    def contract(self, address, abi):
        return self.w3.eth.contract(
            address=Web3.to_checksum_address(address),
            abi=abi,
        )

    def get_usdc_balance(self, address: EthAddress | None = None) -> float:
        """
        Get the usdc balance of the given address.

        If no address is given, the balance of the proxy account corresponding to
        the private key is returned (i.e. Polymarket balance).
        Explicitly passing the proxy address is faster due to only one contract function call.
        """
        if address is None:
            address = self.exchange.functions.getPolyProxyWalletAddress(self.account.address).call()
        balance_res = self.usdc.functions.balanceOf(address).call()
        return float(balance_res / 1e6)

    def get_token_balance(self, token_id: str, address: EthAddress | None = None) -> float:
        """Get the token balance of the given address."""
        if address is None:
            address = self.exchange.functions.getPolyProxyWalletAddress(self.account.address).call()
        balance_res = self.conditional_tokens.functions.balanceOf(address, int(token_id)).call()
        return float(balance_res / 1e6)

    def get_token_complement(self, token_id: str) -> Optional[str]:
        """Get the complement of the given token."""
        try:
            return str(self.neg_risk_exchange.functions.getComplement(int(token_id)).call())
        except ContractCustomError as e:
            if e.args[0] in CUSTOM_ERROR_DICT:
                try:
                    return str(self.exchange.functions.getComplement(int(token_id)).call())
                except ContractCustomError as e2:
                    if e2.args[0] in CUSTOM_ERROR_DICT:
                        msg = f"{CUSTOM_ERROR_DICT[e2.args[0]]}"
                        raise ContractCustomError(
                            msg,
                        ) from e2

    def get_condition_id_neg_risk(self, question_id: Keccak256) -> Keccak256:
        """
        Get the condition id for a given question id.

        Warning: this works for neg risk markets (where the
        outcomeSlotCount is represented by the last two digits of question id). Returns a keccak256 hash of
        the oracle and question id.
        """
        return "0x" + self.neg_risk_adapter.functions.getConditionId(question_id).call().hex()

    def split_position(self, condition_id: Keccak256, amount: int, neg_risk: bool = True):
        """Splits usdc into two complementary positions of equal size."""
        nonce = self.w3.eth.get_transaction_count(self.account.address)
        amount = int(amount * 1e6)

        proxy_txn = {
            "typeCode": 1,
            "to": self.neg_risk_adapter_address if neg_risk else self.conditional_tokens_address,
            "value": 0,
            "data": self._encode_split(condition_id, amount),
        }

        # Send transaction through proxy factory
        txn_data = self.proxy_factory.functions.proxy([proxy_txn]).build_transaction({
            "nonce": nonce,
            "gasPrice": int(1.05 * self.w3.eth.gas_price),
            "gas": 1000000,
            "from": self.account.address,
        })

        # Sign and send transaction
        signed_txn = self.account.sign_transaction(txn_data)
        tx_hash = self.w3.eth.send_raw_transaction(signed_txn.raw_transaction).hex()

        print(f"Txn hash: {tx_hash}")

        # Wait for transaction to be mined
        self.w3.eth.wait_for_transaction_receipt(tx_hash)

        print("Done!")

    def merge_position(self, condition_id: Keccak256, amount: int, neg_risk: bool = True):
        """Merges two complementary positions into usdc."""
        nonce = self.w3.eth.get_transaction_count(self.account.address)
        amount = int(amount * 1e6)

        proxy_txn = {
            "typeCode": 1,
            "to": self.neg_risk_adapter_address if neg_risk else self.conditional_tokens_address,
            "value": 0,
            "data": self._encode_merge(condition_id, amount),
        }

        # Send transaction through proxy factory
        txn_data = self.proxy_factory.functions.proxy([proxy_txn]).build_transaction({
            "nonce": nonce,
            "gasPrice": int(1.05 * self.w3.eth.gas_price),
            "gas": 1000000,
            "from": self.account.address,
        })

        # Sign and send transaction
        signed_txn = self.account.sign_transaction(txn_data)
        tx_hash = self.w3.eth.send_raw_transaction(signed_txn.raw_transaction).hex()

        print(f"Txn hash: {tx_hash}")

        # Wait for transaction to be mined
        self.w3.eth.wait_for_transaction_receipt(tx_hash)

        print("Done!")

    def redeem_position(self, condition_id: Keccak256, amounts: list[float], neg_risk: bool = True):
        """
        Redeem a position into usdc.

        Takes a condition id and a list of sizes in shares [x, y]
        where x is the number of shares of the first outcome
              y is the number of shares of the second outcome.
        """
        nonce = self.w3.eth.get_transaction_count(self.account.address)
        amounts = [int(amount * 1e6) for amount in amounts]

        proxy_txn = {
            "typeCode": 1,
            "to": self.neg_risk_adapter_address if neg_risk else self.conditional_tokens_address,
            "value": 0,
            "data": self._encode_redeem_neg_risk(condition_id, amounts) if neg_risk else self._encode_redeem(condition_id),
        }

        # Send transaction through proxy factory
        txn_data = self.proxy_factory.functions.proxy([proxy_txn]).build_transaction({
            "nonce": nonce,
            "gasPrice": int(1.05 * self.w3.eth.gas_price),
            "gas": 1000000,
            "from": self.account.address,
        })

        # Sign and send transaction
        signed_txn = self.account.sign_transaction(txn_data)
        tx_hash = self.w3.eth.send_raw_transaction(signed_txn.raw_transaction).hex()

        print(f"Txn hash: {tx_hash}")

        # Wait for transaction to be mined
        self.w3.eth.wait_for_transaction_receipt(tx_hash)

        print("Done!")

    def convert_positions(self, question_ids: list[Keccak256], neg_risk_market_id: Keccak256, amount: int):
        nonce = self.w3.eth.get_transaction_count(self.account.address)
        amount = int(amount * 1e6)

        proxy_txn = {
            "typeCode": 1,
            "to": self.neg_risk_adapter_address,
            "value": 0,
            "data": self._encode_convert(neg_risk_market_id, get_index_set(question_ids), amount),
        }

        txn_data = self.proxy_factory.functions.proxy([proxy_txn]).build_transaction({
            "nonce": nonce,
            "gasPrice": int(1.05 * self.w3.eth.gas_price),
            "gas": 1000000,
            "from": self.account.address,
        })

        signed_txn = self.account.sign_transaction(txn_data)
        tx_hash = self.w3.eth.send_raw_transaction(signed_txn.raw_transaction).hex()

        print(f"Txn hash: {tx_hash}")

        # Wait for transaction to be mined
        self.w3.eth.wait_for_transaction_receipt(tx_hash)

        print("Done!")



import json
import logging
from datetime import UTC, datetime
from typing import Literal, Optional
from urllib.parse import urljoin

import httpx
from httpx import HTTPStatusError
from py_order_utils.model import SignedOrder

from ..types.clob_types import (
    ApiCreds,
    BidAsk,
    BookParams,
    ClobMarket,
    CreateOrderOptions,
    DailyEarnedReward,
    MarketOrderArgs,
    MarketRewards,
    Midpoint,
    OpenOrder,
    OrderArgs,
    OrderBookSummary,
    OrderCancelResponse,
    OrderPostResponse,
    OrderType,
    PaginatedResponse,
    PartialCreateOrderOptions,
    PolygonTrade,
    PolymarketRewardItem,
    PostOrdersArgs,
    Price,
    PriceHistory,
    RequestArgs,
    Spread,
    TickSize,
    TokenBidAskDict,
    TokenValueDict,
)
from ..types.common import EthAddress, Keccak256
from ..utilities.constants import END_CURSOR, POLYGON
from ..utilities.endpoints import (
    ARE_ORDERS_SCORING,
    CANCEL,
    CANCEL_ALL,
    CANCEL_ORDERS,
    CREATE_API_KEY,
    DELETE_API_KEY,
    DERIVE_API_KEY,
    GET_API_KEYS,
    GET_FEE_RATE,
    GET_LAST_TRADE_PRICE,
    GET_LAST_TRADES_PRICES,
    GET_MARKET,
    GET_MARKETS,
    GET_NEG_RISK,
    GET_ORDER_BOOK,
    GET_ORDER_BOOKS,
    GET_PRICES,
    GET_SPREAD,
    GET_SPREADS,
    GET_TICK_SIZE,
    IS_ORDER_SCORING,
    MID_POINT,
    MID_POINTS,
    ORDERS,
    POST_ORDER,
    POST_ORDERS,
    PRICE,
    TIME,
    TRADES,
)
from ..utilities.exceptions import (
    InvalidFeeRateError,
    InvalidPriceError,
    InvalidTickSizeError,
    LiquidityError,
    MissingOrderbookError,
)
from ..utilities.headers import create_level_1_headers, create_level_2_headers
from ..utilities.order_builder.builder import OrderBuilder
from ..utilities.order_builder.helpers import (
    is_tick_size_smaller,
    order_to_json,
    price_valid,
)
from ..utilities.signing.signer import Signer

logger = logging.getLogger(__name__)

class PolymarketClobClient:
    def __init__(
            self,
            private_key: str,
            proxy_address: EthAddress,
            creds: Optional[ApiCreds] = None,
            chain_id: Literal[137, 80002] = POLYGON,
            signature_type: Literal[0, 1, 2] = 1,
            # 0 - EOA wallet, 1 - Proxy wallet, 2 - Gnosis Safe wallet
    ):
        self.proxy_address = proxy_address
        self.client = httpx.Client(http2=True, timeout=30.0)
        self.async_client = httpx.AsyncClient(http2=True, timeout=30.0)
        self.base_url: str = "https://clob.polymarket.com"
        self.signer = Signer(private_key=private_key, chain_id=chain_id)
        self.builder = OrderBuilder(
            signer=self.signer,
            sig_type=signature_type,
            funder=proxy_address,
        )
        self.creds = creds if creds else self.create_or_derive_api_creds()

        # local cache
        self.__tick_sizes = {}
        self.__neg_risk = {}
        self.__fee_rates = {}

    def _build_url(self, endpoint: str) -> str:
        return urljoin(self.base_url, endpoint)

    def get_ok(self) -> str:
        response = self.client.get(self.base_url)
        response.raise_for_status()
        return response.json()

    def create_api_creds(self, nonce: Optional[int] = None) -> ApiCreds:
        headers = create_level_1_headers(self.signer, nonce)
        response = self.client.post(self._build_url(CREATE_API_KEY), headers=headers)
        response.raise_for_status()
        return ApiCreds(**response.json())

    def derive_api_key(self, nonce: Optional[int] = None) -> ApiCreds:
        headers = create_level_1_headers(self.signer, nonce)
        response = self.client.get(self._build_url(DERIVE_API_KEY), headers=headers)
        response.raise_for_status()
        return ApiCreds(**response.json())

    def create_or_derive_api_creds(self, nonce: Optional[int] = None) -> ApiCreds:
        try:
            return self.create_api_creds(nonce)
        except HTTPStatusError:
            return self.derive_api_key(nonce)

    def set_api_creds(self, creds: ApiCreds) -> None:
        self.creds = creds

    def get_api_keys(self) -> dict:
        request_args = RequestArgs(method="GET", request_path=GET_API_KEYS)
        headers = create_level_2_headers(self.signer, self.creds, request_args)
        response = self.client.get(self._build_url(GET_API_KEYS), headers=headers)
        response.raise_for_status()
        return response.json()

    def delete_api_keys(self) -> Literal["OK"]:
        request_args = RequestArgs(method="DELETE", request_path=DELETE_API_KEY)
        headers = create_level_2_headers(self.signer, self.creds, request_args)
        response = self.client.delete(self._build_url(DELETE_API_KEY), headers=headers)
        response.raise_for_status()
        return response.json()

    def get_utc_time(self) -> datetime:
        # parse server timestamp into utc datetime
        response = self.client.get(self._build_url(TIME))
        response.raise_for_status()
        return datetime.fromtimestamp(response.json(), tz=UTC)

    def get_tick_size(self, token_id: str) -> TickSize:
        if token_id in self.__tick_sizes:
            return self.__tick_sizes[token_id]

        params = {"token_id": token_id}
        response = self.client.get(self._build_url(GET_TICK_SIZE), params=params)
        response.raise_for_status()
        self.__tick_sizes[token_id] = str(response.json()["minimum_tick_size"])

        return self.__tick_sizes[token_id]

    def get_neg_risk(self, token_id: str) -> bool:
        if token_id in self.__neg_risk:
            return self.__neg_risk[token_id]

        params = {"token_id": token_id}
        response = self.client.get(self._build_url(GET_NEG_RISK), params=params)
        response.raise_for_status()
        self.__neg_risk[token_id] = response.json()["neg_risk"]

        return self.__neg_risk[token_id]

    def get_fee_rate_bps(self, token_id: str) -> int:
        if token_id in self.__fee_rates:
            return self.__fee_rates[token_id]

        params = {"token_id": token_id}
        response = self.client.get(self._build_url(GET_FEE_RATE), params=params)
        response.raise_for_status()
        fee_rate = response.json().get("base_fee") or 0
        self.__fee_rates[token_id] = fee_rate

        return fee_rate

    def __resolve_tick_size(
            self, token_id: str, tick_size: TickSize = None,
    ) -> TickSize:
        min_tick_size = self.get_tick_size(token_id)
        if tick_size is not None:
            if is_tick_size_smaller(tick_size, min_tick_size):
                msg = f"invalid tick size ({tick_size!s}), minimum for the market is {min_tick_size!s}"
                raise InvalidTickSizeError(msg)
        else:
            tick_size = min_tick_size
        return tick_size

    def __resolve_fee_rate(
            self, token_id: str, user_fee_rate: Optional[int] = None,
    ) -> int:
        market_fee_rate_bps = self.get_fee_rate_bps(token_id)
        # If both fee rate on the market and the user supplied fee rate are non-zero, validate that they match
        # else return the market fee rate
        if market_fee_rate_bps > 0 and user_fee_rate is not None and user_fee_rate > 0 and user_fee_rate != market_fee_rate_bps:
            msg = f"invalid user provided fee rate: ({user_fee_rate}), fee rate for the market must be {market_fee_rate_bps}"
            raise InvalidFeeRateError(msg)
        return market_fee_rate_bps

    def get_midpoint(self, token_id: str) -> Midpoint:
        """Get the mid-market price for the given token."""
        params = {"token_id": token_id}
        response = self.client.get(self._build_url(MID_POINT), params=params)
        response.raise_for_status()
        return Midpoint(token_id=token_id, value=float(response.json()["mid"]))

    def get_midpoints(self, token_ids: list[str]) -> dict:
        """Get the mid-market prices for a set of tokens."""
        data = [{"token_id": token_id} for token_id in token_ids]
        response = self.client.post(self._build_url(MID_POINTS), json=data)
        response.raise_for_status()
        return TokenValueDict(**response.json()).root

    def get_spread(self, token_id: str) -> Spread:
        """Get the spread for the given token."""
        params = {"token_id": token_id}
        response = self.client.get(self._build_url(GET_SPREAD), params=params)
        response.raise_for_status()
        return Spread(token_id=token_id, value=float(response.json()["mid"]))

    def get_spreads(self, token_ids: list[str]) -> dict:
        """Get the spreads for a set of tokens."""
        data = [{"token_id": token_id} for token_id in token_ids]
        response = self.client.post(self._build_url(GET_SPREADS), json=data)
        response.raise_for_status()
        return TokenValueDict(**response.json()).root

    def get_price(self, token_id: str, side: Literal["BUY", "SELL"]) -> Price:
        """Get the market price for the given token and side."""
        params = {"token_id": token_id, "side": side}
        response = self.client.get(self._build_url(PRICE), params=params)
        response.raise_for_status()
        return Price(**response.json(), token_id=token_id, side=side)

    def get_prices(self, params: list[BookParams]) -> dict[str, BidAsk]:
        """Get the market prices for a set of tokens and sides."""
        data = [{"token_id": param.token_id, "side": param.side} for param in params]
        response = self.client.post(self._build_url(GET_PRICES), json=data)
        response.raise_for_status()
        return TokenBidAskDict(**response.json()).root

    def get_last_trade_price(self, token_id) -> Price:
        """Fetches the last trade price for a token_id."""
        params = {"token_id": token_id}
        response = self.client.get(self._build_url(GET_LAST_TRADE_PRICE), params=params)
        response.raise_for_status()
        return Price(**response.json(), token_id=token_id)

    def get_last_trades_prices(self, token_ids: list[str]) -> list[Price]:
        """Fetches the last trades prices for a set of token ids."""
        body = [{"token_id": token_id} for token_id in token_ids]
        response = self.client.post(self._build_url(GET_LAST_TRADES_PRICES), json=body)
        response.raise_for_status()
        return [Price(**price) for price in response.json()]

    def get_order_book(self, token_id) -> OrderBookSummary:
        """Get the orderbook for the given token."""
        params = {"token_id": token_id}
        response = self.client.get(self._build_url(GET_ORDER_BOOK), params=params)
        response.raise_for_status()
        return OrderBookSummary(**response.json())

    def get_order_books(self, token_ids: list[str]) -> list[OrderBookSummary]:
        """Get the orderbook for a set of tokens."""
        body = [{"token_id": token_id} for token_id in token_ids]
        response = self.client.post(self._build_url(GET_ORDER_BOOKS), json=body)
        response.raise_for_status()
        return [OrderBookSummary(**obs) for obs in response.json()]

    async def get_order_books_async(self, token_ids: list[str]) -> list[OrderBookSummary]:
        """Get the orderbook for a set of tokens asynchronously."""
        body = [{"token_id": token_id} for token_id in token_ids]
        response = await self.async_client.post(self._build_url(GET_ORDER_BOOKS), json=body)
        response.raise_for_status()
        return [OrderBookSummary(**obs) for obs in response.json()]

    def get_market(self, condition_id) -> ClobMarket:
        """Get a ClobMarket by condition_id."""
        response = self.client.get(self._build_url(GET_MARKET + condition_id))
        response.raise_for_status()
        return ClobMarket(**response.json())

    def get_markets(self, next_cursor="MA==")  -> PaginatedResponse[ClobMarket]:
        """Get paginated ClobMarkets."""
        params = {"next_cursor": next_cursor}
        response = self.client.get(self._build_url(GET_MARKETS), params=params)
        response.raise_for_status()
        return PaginatedResponse[ClobMarket](**response.json())

    def get_all_markets(self, next_cursor="MA==") -> list[ClobMarket]:
        """Recursively fetch all ClobMarkets using pagination."""
        # Base case: Stop recursion if next_cursor indicates the last page
        if next_cursor == "LTE=":
            print("Reached the last page of markets.")
            return []

        # Fetch current page of markets
        paginated_response = self.get_markets(next_cursor=next_cursor)

        # Collect current page data
        current_markets = paginated_response.data

        # Recursively fetch remaining pages
        next_page_markets = self.get_all_markets(
            next_cursor=paginated_response.next_cursor,
        )

        # Combine current page data with data from subsequent pages
        return current_markets + next_page_markets

    def get_recent_history(
            self,
            token_id: str,
            interval: Optional[Literal["1d", "6h", "1h"]] = "1d",
            fidelity: int = 1,  # resolution in minutes
    ) -> PriceHistory:
        """Get the recent price history of a token (up to now) - 1h, 6h, 1d."""
        if fidelity < 1:
            msg = f"invalid filters: minimum 'fidelity' for '{interval}' range is 1"
            raise ValueError(msg)

        params = {
            "market": token_id,
            "interval": interval,
            "fidelity": fidelity,
        }
        response = self.client.get(self._build_url("/prices-history"), params=params)
        response.raise_for_status()
        return PriceHistory(**response.json(), token_id=token_id)

    def get_history(
            self,
            token_id: str,
            start_time: Optional[datetime] = None,
            end_time: Optional[datetime] = None,
            interval: Optional[Literal["max", "1m", "1w"]] = "max",
            fidelity: Optional[int] = 2,  # resolution in minutes
    ) -> PriceHistory:
        """Get the price history of a token between selected dates - 1m, 1w, max."""
        min_fidelities = {"1m": 10, "1w": 5, "max": 2}

        if fidelity < min_fidelities[interval]:
            msg = f"invalid filters: minimum 'fidelity' for '{interval}' range is {min_fidelities[interval]}"
            raise ValueError(msg)

        if start_time is None and end_time is None:
            msg = "At least one of 'start_time' or 'end_time' must be provided."
            raise ValueError(msg)

        # Default values for timestamps if one is not provided

        if start_time is None:
            start_time = datetime(2020, 1, 1, tzinfo=UTC)  # Default start time
        if end_time is None:
            end_time = datetime.now(UTC)  # Default end time

        params = {
            "market": token_id,
            "startTs": int(start_time.timestamp()),
            "endTs": int(end_time.timestamp()),
            "interval": interval,
            "fidelity": fidelity,
        }
        response = self.client.get(self._build_url("/prices-history"), params=params)
        response.raise_for_status()
        return PriceHistory(**response.json(), token_id=token_id)

    def get_orders(self, order_id: Optional[str] = None, condition_id: Optional[Keccak256] = None, token_id: Optional[str] = None, next_cursor: str ="MA==") -> list[OpenOrder]:
        """Gets your active orders, filtered by order_id, condition_id, token_id."""
        params = {}
        if order_id:
            params["id"] = order_id
        if condition_id:
            params["market"] = condition_id
        if token_id:
            params["asset_id"] = token_id

        request_args = RequestArgs(method="GET", request_path=ORDERS)
        headers = create_level_2_headers(self.signer, self.creds, request_args)

        results = []
        next_cursor = next_cursor if next_cursor is not None else "MA=="
        while next_cursor != END_CURSOR:
            params["next_cursor"] = next_cursor
            response = self.client.get(self._build_url(ORDERS), headers=headers, params=params)
            response.raise_for_status()
            next_cursor = response.json()["next_cursor"]
            results += [OpenOrder(**order) for order in response.json()["data"]]

        return results

    def create_order(self, order_args: OrderArgs, options: Optional[PartialCreateOrderOptions] = None) -> SignedOrder:
        """Creates and signs an order."""
        # add resolve_order_options, or similar
        tick_size = self.__resolve_tick_size(
            order_args.token_id,
            options.tick_size if options else None,
        )

        if not price_valid(order_args.price, tick_size):
            msg = f"price ({order_args.price}), min: {tick_size} - max: {1 - float(tick_size)}"
            raise InvalidPriceError(msg)


        neg_risk = (
            options.neg_risk
            if options and options.neg_risk
            else self.get_neg_risk(order_args.token_id)
        )

        # fee rate
        fee_rate_bps = self.__resolve_fee_rate(order_args.token_id, order_args.fee_rate_bps)
        order_args.fee_rate_bps = fee_rate_bps

        return self.builder.create_order(
            order_args,
            CreateOrderOptions(
                tick_size=tick_size,
                neg_risk=neg_risk,
            ),
        )

    def post_order(self, order: SignedOrder, order_type: OrderType = OrderType.GTC) -> Optional[OrderPostResponse]:
        """Posts a SignedOrder."""
        body = order_to_json(order, self.creds.key, order_type)
        headers = create_level_2_headers(
            self.signer,
            self.creds,
            RequestArgs(method="POST", request_path=POST_ORDER, body=body),
        )

        try:
            response = self.client.post(
                self._build_url("/order"),
                headers=headers,
                content=json.dumps(body).encode("utf-8"),
            )
            response.raise_for_status()
            return OrderPostResponse(**response.json())
        except httpx.HTTPStatusError as exc:
            msg = f"Client Error '{exc.response.status_code} {exc.response.reason_phrase}' while posting order"
            logger.warning(msg)
            error_json = exc.response.json()
            print("Details:", error_json["error"])

    def create_and_post_order(self, order_args: OrderArgs, options: Optional[PartialCreateOrderOptions] = None, order_type: OrderType = OrderType.GTC) -> OrderPostResponse:
        """Utility function to create and publish an order."""
        order = self.create_order(order_args, options)
        return self.post_order(order=order, order_type=order_type)

    def post_orders(self, args: list[PostOrdersArgs]):
        """Posts multiple SignedOrders at once."""
        body = [order_to_json(arg.order, self.creds.key, arg.order_type) for arg in args]
        headers = create_level_2_headers(
            self.signer,
            self.creds,
            RequestArgs(method="POST", request_path=POST_ORDERS, body=body),
        )

        try:
            response = self.client.post(
                self._build_url("/orders"),
                headers=headers,
                content=json.dumps(body).encode("utf-8"),
            )
            response.raise_for_status()
            order_responses = []
            for index, item in enumerate(response.json()):
                resp = OrderPostResponse(**item)
                order_responses.append(resp)
                if resp.error_msg:
                    msg = (f"Error posting order in position {index} \n"
                           f"Details: {resp.error_msg}")
                    logger.warning(msg)
        except httpx.HTTPStatusError as exc:
            msg = f"Client Error '{exc.response.status_code} {exc.response.reason_phrase}' while posting order"
            logger.warning(msg)
            error_json = exc.response.json()
            print("Details:", error_json["error"])
        else:
            return order_responses

    def create_and_post_orders(self, args: list[OrderArgs], order_types: list[OrderType]) -> list[OrderPostResponse]:
        """Utility function to create and publish multiple orders at once."""
        return self.post_orders(
            [PostOrdersArgs(order=self.create_order(order_args),
                            order_type=order_type)
             for order_args, order_type in zip(args, order_types, strict=True)],
        )

    def calculate_market_price(self, token_id: str, side: str, amount: float, order_type: OrderType) -> float:
        """Calculates the matching price considering an amount and the current orderbook."""
        book = self.get_order_book(token_id)
        if book is None:
            msg = "Order book is None"
            raise MissingOrderbookError(msg)
        if side == "BUY":
            if book.asks is None:
                msg = "No ask orders available"
                raise LiquidityError(msg)
            return self.builder.calculate_buy_market_price(
                book.asks, amount, order_type,
            )
        if side == "SELL":
            if book.bids is None:
                msg = "No bid orders available"
                raise LiquidityError(msg)
            return self.builder.calculate_sell_market_price(
                book.bids, amount, order_type,
            )
        msg = 'Side must be "BUY" or "SELL"'
        raise ValueError(msg)

    def create_market_order(self, order_args: MarketOrderArgs, options: Optional[PartialCreateOrderOptions] = None):
        """Creates and signs a market order."""
        tick_size = self.__resolve_tick_size(
            order_args.token_id,
            options.tick_size if options else None,
        )

        if order_args.price is None or order_args.price <= 0:
            order_args.price = self.calculate_market_price(
                order_args.token_id,
                order_args.side,
                order_args.amount,
                order_args.order_type,
            )

        if not price_valid(order_args.price, tick_size):
            msg = f"price ({order_args.price}), min: {tick_size} - max: {1 - float(tick_size)}"
            raise InvalidPriceError(msg)

        neg_risk = (
            options.neg_risk
            if options and options.neg_risk
            else self.get_neg_risk(order_args.token_id)
        )

        # fee rate
        fee_rate_bps = self.__resolve_fee_rate(order_args.token_id, order_args.fee_rate_bps)
        order_args.fee_rate_bps = fee_rate_bps

        return self.builder.create_market_order(
            order_args,
            CreateOrderOptions(
                tick_size=tick_size,
                neg_risk=neg_risk,
            ),
        )

    def create_and_post_market_order(
            self,
            order_args: MarketOrderArgs,
            options: Optional[PartialCreateOrderOptions] = None,
            order_type: OrderType = OrderType.FOK,
    ) -> OrderPostResponse:
        """Utility function to create and publish a market order."""
        order = self.create_market_order(order_args, options)
        return self.post_order(order=order, order_type=order_type)

    def cancel_order(self, order_id: Keccak256) -> OrderCancelResponse:
        """Cancels an order."""
        body = {"orderID": order_id}

        request_args = RequestArgs(method="DELETE", request_path=CANCEL, body=body)
        headers = create_level_2_headers(self.signer, self.creds, request_args)

        response = self.client.request("DELETE", self._build_url(CANCEL), headers=headers, data=json.dumps(body).encode("utf-8"))
        response.raise_for_status()
        return OrderCancelResponse(**response.json())

    def cancel_orders(self, order_ids: list[Keccak256]) -> OrderCancelResponse:
        """Cancels orders."""
        body = order_ids

        request_args = RequestArgs(
            method="DELETE", request_path=CANCEL_ORDERS, body=body,
        )
        headers = create_level_2_headers(self.signer, self.creds, request_args)

        response = self.client.request("DELETE", self._build_url(CANCEL_ORDERS), headers=headers, data=json.dumps(body).encode("utf-8"))
        response.raise_for_status()
        return OrderCancelResponse(**response.json())

    def cancel_all(self) -> OrderCancelResponse:
        """Cancels all available orders for the user."""
        request_args = RequestArgs(method="DELETE", request_path=CANCEL_ALL)
        headers = create_level_2_headers(self.signer, self.creds, request_args)

        response = self.client.delete(self._build_url(CANCEL_ALL), headers=headers)
        response.raise_for_status()
        return OrderCancelResponse(**response.json())

    def is_order_scoring(self, order_id: Keccak256) -> bool:
        """Check if the order is currently scoring."""
        request_args = RequestArgs(method="GET", request_path=IS_ORDER_SCORING)
        headers = create_level_2_headers(self.signer, self.creds, request_args)

        response = self.client.get(self._build_url(IS_ORDER_SCORING), headers=headers, params={"order_id": order_id})
        response.raise_for_status()
        return response.json()["scoring"]

    def are_orders_scoring(self, order_ids: list[Keccak256]) -> dict[Keccak256, bool]:
        """Check if the orders are currently scoring."""
        body = order_ids
        request_args = RequestArgs(
            method="POST", request_path=ARE_ORDERS_SCORING, body=body,
        )
        headers = create_level_2_headers(self.signer, self.creds, request_args)
        headers["Content-Type"] = "application/json"

        response = self.client.post(self._build_url(ARE_ORDERS_SCORING), headers=headers, json=body)
        response.raise_for_status()
        return response.json()

    def get_market_rewards(self, condition_id: Keccak256) -> MarketRewards:
        """
        Get the MarketRewards for a given market (condition_id).

        - metadata, tokens, max_spread, min_size, rewards_config, market_competitiveness.
        """
        request_args = RequestArgs(method="GET", request_path="/rewards/markets/")
        headers = create_level_2_headers(self.signer, self.creds, request_args)

        response = self.client.get(self._build_url("/rewards/markets/" + condition_id), headers=headers)
        response.raise_for_status()
        return next(MarketRewards(**market) for market in response.json()["data"])

    def get_trades(
            self,
            condition_id: Optional[Keccak256] = None,
            token_id: Optional[str] = None,
            trade_id: Optional[str] = None,
            before: Optional[datetime] = None,
            after: Optional[datetime] = None,
            proxy_address: Optional[int] = None,
            next_cursor="MA==") -> list[PolygonTrade]:
        """Fetches the trade history for a user."""
        params = {}
        if condition_id:
            params["market"] = condition_id
        if token_id:
            params["asset_id"] = token_id
        if trade_id:
            params["id"] = trade_id
        if before:
            params["before"] = int(before.replace(microsecond=0).timestamp())
        if after:
            params["after"] = int(after.replace(microsecond=0).timestamp())
        if proxy_address:
            params["maker_address"] = proxy_address

        request_args = RequestArgs(method="GET", request_path=TRADES)
        headers = create_level_2_headers(self.signer, self.creds, request_args)

        results = []
        next_cursor = next_cursor if next_cursor is not None else "MA=="
        while next_cursor != END_CURSOR:
            params["next_cursor"] = next_cursor
            response =  self.client.get(self._build_url(TRADES), headers=headers, params=params)
            response.raise_for_status()
            next_cursor = response.json()["next_cursor"]
            results += [PolygonTrade(**trade) for trade in response.json()["data"]]

        return results

    def get_total_rewards(self, date: Optional[datetime] = None) -> DailyEarnedReward:
        """Get the total rewards earned on a given date (seems to only hold the 6 most recent data points)."""
        if date is None:
            date = datetime.now(UTC)
        params = {
            "authenticationType": "magic",
            "date": f"{date.strftime("%Y-%m-%d")}",
        }

        request_args = RequestArgs(method="GET", request_path="/rewards/user/total")
        headers = create_level_2_headers(self.signer, self.creds, request_args)
        params["l2Headers"] = json.dumps(headers)

        response = self.client.get("https://polymarket.com/api/rewards/totalEarnings", params=params)
        response.raise_for_status()
        if response.json():
            return DailyEarnedReward(**response.json()[0])
        return DailyEarnedReward(
            date=date,
            asset_address="0x2791Bca1f2de4661ED88A30C99A7a9449Aa84174",
            maker_address=self.proxy_address,
            earnings=0.0,
            asset_rate=0.0,
        )

    def get_reward_markets(
            self,
            sort_by: Optional[Literal["market", "max_spread", "min_size", "rate_per_day", "spread", "price", "earnings", "earning_percentage"]] = "market",
            sort_direction: Optional[Literal["ASC", "DESC"]] = None,
            query: Optional[str] = None,
            show_favorites: bool = False,
    ) -> list[PolymarketRewardItem]:
        """
        Get all polymarket.com/rewards items, sorted by different criteria.

         - market start date ("market") - TODO confirm this
         - max spread for rewards in usdc
         - min size for rewards in shares
         - reward rate per day in usdc
         - current spread of a market
         - current price of a market
         - your daily earnings on a market - only need auth for these last two
         - your current earning percentage on a market.
        """
        results = []
        desc = {"ASC": False, "DESC": True}
        params = {
            "authenticationType": "magic",
            "showFavorites": show_favorites,
        }
        if sort_by:
            params["orderBy"] = sort_by
        if query:
            params["query"] = query
            params["desc"] = False
        if sort_direction:
            params["desc"] = desc[sort_direction]

        request_args = RequestArgs(method="GET", request_path="/rewards/user/markets")
        headers = create_level_2_headers(self.signer, self.creds, request_args)
        params["l2Headers"] = json.dumps(headers)

        next_cursor = "MA=="
        while next_cursor != END_CURSOR:
            params["nextCursor"] = next_cursor
            response = self.client.get("https://polymarket.com/api/rewards/markets", params=params)
            # can probably use clob/rewards/user/markets here but haven't figure out auth
            response.raise_for_status()
            next_cursor = response.json()["next_cursor"]
            results += [PolymarketRewardItem(**reward) for reward in response.json()["data"]]

        return results

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.client.close()



from collections.abc import Callable
from json import JSONDecodeError
from typing import Any, Optional

from lomond import WebSocket
from lomond.persist import persist
from pydantic import ValidationError

from ..types.clob_types import ApiCreds
from ..types.websockets_types import (
    ActivityOrderMatchEvent,
    ActivityTradeEvent,
    CommentEvent,
    CryptoPriceSubscribeEvent,
    CryptoPriceUpdateEvent,
    LastTradePriceEvent,
    LiveDataLastTradePriceEvent,
    LiveDataOrderBookSummaryEvent,
    LiveDataOrderEvent,
    LiveDataPriceChangeEvent,
    LiveDataTickSizeChangeEvent,
    LiveDataTradeEvent,
    MarketStatusChangeEvent,
    OrderBookSummaryEvent,
    OrderEvent,
    PriceChangeEvent,
    QuoteEvent,
    ReactionEvent,
    RequestEvent,
    TickSizeChangeEvent,
    TradeEvent,
)
from ..utilities.exceptions import AuthenticationRequiredError


def _process_market_event(event):
    try:
        message = event.json
        if isinstance(message, list):
            for item in message:
                try:
                    print(OrderBookSummaryEvent(**item), "\n")
                except ValidationError as e:
                    print(item.text)
                    print(e.errors())
            return
        match message["event_type"]:
            case "book":
                print(OrderBookSummaryEvent(**message), "\n")
            case "price_change":
                print(PriceChangeEvent(**message), "\n")
            case "tick_size_change":
                print(TickSizeChangeEvent(**message), "\n")
            case "last_trade_price":
                print(LastTradePriceEvent(**message), "\n")
            case _:
                print(message)
    except JSONDecodeError:
        print(event.text)
    except ValidationError as e:
        print(e.errors())
        print(event.json)

def _process_user_event(event):
    try:
        message = event.json
        match message["event_type"]:
            case "order":
                print(OrderEvent(**message), "\n")
            case "trade":
                print(TradeEvent(**message), "\n")
    except JSONDecodeError:
        print(event.text)
    except ValidationError as e:
        print(event.text)
        print(e.errors(), "\n")

def _process_live_data_event(event):
    try:
        message = event.json
        match message["type"]:
            case "trades":
                print(ActivityTradeEvent(**message), "\n")
            case "orders_matched":
                print(ActivityOrderMatchEvent(**message), "\n")
            case "comment_created" | "comment_removed":
                print(CommentEvent(**message), "\n")
            case "reaction_created" | "reaction_removed":
                print(ReactionEvent(**message), "\n")
            case "request_created" | "request_edited" | "request_canceled" | "request_expired":
                print(RequestEvent(**message), "\n")
            case "quote_created" | "quote_edited" | "quote_canceled" | "quote_expired":
                print(QuoteEvent(**message), "\n")
            case "subscribe":
                print(CryptoPriceSubscribeEvent(**message), "\n")
            case "update":
                print(CryptoPriceUpdateEvent(**message), "\n")
            case "agg_orderbook":
                print(LiveDataOrderBookSummaryEvent(**message), "\n")
            case "price_change":
                print(LiveDataPriceChangeEvent(**message), "\n")
            case "last_trade_price":
                print(LiveDataLastTradePriceEvent(**message), "\n")
            case "tick_size_change":
                print(LiveDataTickSizeChangeEvent(**message), "\n")
            case "market_created" | "market_resolved":
                print(MarketStatusChangeEvent(**message), "\n")
            case "order":
                print(LiveDataOrderEvent(**message), "\n")
            case "trade":
                print(LiveDataTradeEvent(**message), "\n")
            case _:
                print(message)
    except JSONDecodeError:
        print(event.text)
    except ValidationError as e:
        print(e.errors(), "\n")
        print(event.text)

class PolymarketWebsocketsClient:
    def __init__(self):
        self.url_market = "wss://ws-subscriptions-clob.polymarket.com/ws/market"
        self.url_user = "wss://ws-subscriptions-clob.polymarket.com/ws/user"
        self.url_live_data = "wss://ws-live-data.polymarket.com"

    def market_socket(self, token_ids: list[str], process_event: Callable = _process_market_event):
        """
        Connect to the market websocket and subscribe to market events for specific token IDs.

        Args:
            token_ids: List of token IDs to subscribe to
            process_event: Callback function to process received events

        """
        websocket = WebSocket(self.url_market)

        for event in persist(websocket):  # persist automatically reconnects
            if event.name == "ready":
                websocket.send_json(
                    assets_ids=token_ids,
                )
            elif event.name == "text":
                process_event(event)

    def user_socket(self, creds: ApiCreds, process_event: Callable = _process_user_event):
        """
        Connect to the user websocket and subscribe to user events.

        Args:
            creds: API credentials for authentication
            process_event: Callback function to process received events

        """
        websocket = WebSocket(self.url_user)

        for event in persist(websocket):
            if event.name == "ready":
                websocket.send_json(
                    auth=creds.model_dump(by_alias=True),
                )
            elif event.name == "text":
                process_event(event)

    def live_data_socket(self, subscriptions: list[dict[str, Any]], process_event: Callable = _process_live_data_event, creds: Optional[ApiCreds] = None):
        # info on how to subscribe found at https://github.com/Polymarket/real-time-data-client?tab=readme-ov-file#subscribe
        """
        Connect to the live data websocket and subscribe to specified events.

        Args:
            subscriptions: List of subscription configurations
            process_event: Callback function to process received events
            creds: ApiCreds for authentication if subscribing to clob_user topic

        """
        websocket = WebSocket(self.url_live_data)

        needs_auth = any(sub.get("topic") == "clob_user" for sub in subscriptions)
        if needs_auth and creds is None:
            msg = "ApiCreds credentials are required for the clob_user topic subscriptions"
            raise AuthenticationRequiredError(msg)

        for event in persist(websocket):
            if event.name == "ready":
                if needs_auth:
                    subscriptions_with_creds = []
                    for sub in subscriptions:
                        if sub.get("topic") == "clob_user":
                            sub_copy = sub.copy()
                            sub_copy["clob_auth"] = creds.model_dump()
                            subscriptions_with_creds.append(sub_copy)
                        else:
                            subscriptions_with_creds.append(sub)
                    subscriptions = subscriptions_with_creds

                payload = {
                    "action": "subscribe",
                    "subscriptions": subscriptions,
                }

                websocket.send_json(**payload)

            elif event.name == "text":
                process_event(event)



from datetime import UTC, datetime
from typing import Literal

from pydantic import BaseModel, Field, field_validator

from .common import EmptyString, EthAddress, Keccak256


class Position(BaseModel):
    # User identification
    proxy_wallet: EthAddress = Field(alias="proxyWallet")

    # Asset information
    token_id: str = Field(alias="asset")
    complementary_token_id: str = Field(alias="oppositeAsset")
    condition_id: Keccak256 = Field(alias="conditionId")
    outcome: str
    complementary_outcome: str = Field(alias="oppositeOutcome")
    outcome_index: int = Field(alias="outcomeIndex")

    # Position details
    size: float
    avg_price: float = Field(alias="avgPrice")
    current_price: float = Field(alias="curPrice")
    redeemable: bool

    # Financial metrics
    initial_value: float = Field(alias="initialValue")
    current_value: float = Field(alias="currentValue")
    cash_pnl: float = Field(alias="cashPnl")
    percent_pnl: float = Field(alias="percentPnl")
    total_bought: float = Field(alias="totalBought")
    realized_pnl: float = Field(alias="realizedPnl")
    percent_realized_pnl: float = Field(alias="percentRealizedPnl")

    # Event information
    title: str
    slug: str
    icon: str
    event_slug: str = Field(alias="eventSlug")
    end_date: datetime = Field(alias="endDate")
    negative_risk: bool = Field(alias="negativeRisk")

    @field_validator("end_date", mode="before")
    def handle_empty_end_date(cls, v):
        if v == "":
            return datetime(2099,12,31, tzinfo=UTC)
        return v


class Trade(BaseModel):
    # User identification
    proxy_wallet: EthAddress = Field(alias="proxyWallet")

    # Trade details
    side: Literal["BUY", "SELL"]
    token_id: str = Field(alias="asset")
    condition_id: Keccak256 = Field(alias="conditionId")
    size: float
    price: float
    timestamp: datetime

    # Event information
    title: str
    slug: str
    icon: str
    event_slug: str = Field(alias="eventSlug")
    outcome: str
    outcome_index: int = Field(alias="outcomeIndex")

    # User profile
    name: str
    pseudonym: str
    bio: str
    profile_image: str = Field(alias="profileImage")
    profile_image_optimized: str = Field(alias="profileImageOptimized")

    # Transaction information
    transaction_hash: Keccak256 = Field(alias="transactionHash")


class Activity(BaseModel):
    # User identification
    proxy_wallet: EthAddress = Field(alias="proxyWallet")

    # Activity details
    timestamp: datetime
    condition_id: Keccak256 | EmptyString = Field(alias="conditionId")
    type: Literal["TRADE", "SPLIT", "MERGE", "REDEEM", "REWARD", "CONVERSION"]
    size: float
    usdc_size: float = Field(alias="usdcSize")
    price: float
    asset: str
    side: str | None
    outcome_index: int = Field(alias="outcomeIndex")

    # Event information
    title: str
    slug: str
    icon: str
    event_slug: str = Field(alias="eventSlug")
    outcome: str

    # User profile
    name: str
    pseudonym: str
    bio: str
    profile_image: str = Field(alias="profileImage")
    profile_image_optimized: str = Field(alias="profileImageOptimized")

    # Transaction information
    transaction_hash: Keccak256 = Field(alias="transactionHash")


class Holder(BaseModel):
    # User identification
    proxy_wallet: EthAddress = Field(alias="proxyWallet")

    # Holder details
    token_id: str = Field(alias="asset")
    amount: float
    outcome_index: int = Field(alias="outcomeIndex")

    # User profile
    name: str
    pseudonym: str
    bio: str
    profile_image: str = Field(alias="profileImage")
    profile_image_optimized: str = Field(alias="profileImageOptimized")
    display_username_public: bool = Field(alias="displayUsernamePublic")


class HolderResponse(BaseModel):
    # Asset information
    token_id: str = Field(alias="token")

    # Holders list
    holders: list[Holder]


class ValueResponse(BaseModel):
    # User identification
    proxy_wallet: EthAddress = Field(alias="proxyWallet")

    # Value information
    value: float

class User(BaseModel):
    proxy_wallet: EthAddress = Field(alias="proxyWallet")
    name: str
    bio: str
    profile_image: str = Field(alias="profileImage")
    profile_image_optimized: str = Field(alias="profileImageOptimized")

class UserMetric(User):
    amount: float
    pseudonym: str

class UserRank(User):
    amount: float
    rank: int



TIME = "/time"
CREATE_API_KEY = "/auth/api-key"
GET_API_KEYS = "/auth/api-keys"
DELETE_API_KEY = "/auth/api-key"
DERIVE_API_KEY = "/auth/derive-api-key"
TRADES = "/data/trades"
GET_ORDER_BOOK = "/book"
GET_ORDER_BOOKS = "/books"
ORDERS = "/data/orders"
POST_ORDER = "/order"
POST_ORDERS = "/orders"
CANCEL = "/order"
CANCEL_ORDERS = "/orders"
CANCEL_ALL = "/cancel-all"
CANCEL_MARKET_ORDERS = "/cancel-market-orders"
MID_POINT = "/midpoint"
MID_POINTS = "/midpoints"
PRICE = "/price"
GET_PRICES = "/prices"
GET_SPREAD = "/spread"
GET_SPREADS = "/spreads"
GET_LAST_TRADE_PRICE = "/last-trade-price"
GET_LAST_TRADES_PRICES = "/last-trades-prices"
GET_NOTIFICATIONS = "/notifications"
DROP_NOTIFICATIONS = "/notifications"
GET_BALANCE_ALLOWANCE = "/balance-allowance"
UPDATE_BALANCE_ALLOWANCE = "/balance-allowance/update"
IS_ORDER_SCORING = "/order-scoring"
ARE_ORDERS_SCORING = "/orders-scoring"
GET_TICK_SIZE = "/tick-size"
GET_NEG_RISK = "/neg-risk"
GET_FEE_RATE = "/fee-rate"
GET_SAMPLING_SIMPLIFIED_MARKETS = "/sampling-simplified-markets"
GET_SAMPLING_MARKETS = "/sampling-markets"
GET_SIMPLIFIED_MARKETS = "/simplified-markets"
GET_MARKETS = "/markets"
GET_MARKET = "/markets/"
GET_MARKET_TRADES_EVENTS = "/live-activity/events/"



def get_market_index(question_id: str) -> int:
    """Extract the market index from a question ID (last 2 hex characters)."""
    return int(question_id[-2:], 16)

def get_index_set(question_ids: list[str]) -> int:
    """Calculate bitwise index set from question IDs."""
    indices = [get_market_index(question_id) for question_id in question_ids]
    return sum(1 << index for index in set(indices))



from datetime import UTC, datetime
from typing import Optional

from ..types.clob_types import ApiCreds, RequestArgs
from .signing.eip712 import sign_clob_auth_message
from .signing.hmac import build_hmac_signature
from .signing.signer import Signer

POLY_ADDRESS = "POLY_ADDRESS"
POLY_SIGNATURE = "POLY_SIGNATURE"
POLY_TIMESTAMP = "POLY_TIMESTAMP"
POLY_NONCE = "POLY_NONCE"
POLY_API_KEY = "POLY_API_KEY"
POLY_PASSPHRASE = "POLY_PASSPHRASE"


def create_level_1_headers(signer: Signer, nonce: Optional[int] = None):
    """Creates Level 1 Poly headers for a request."""
    timestamp = int(datetime.now(tz=UTC).timestamp())

    n = 0
    if nonce is not None:
        n = nonce

    signature = sign_clob_auth_message(signer, timestamp, n)
    headers = {
        POLY_ADDRESS: signer.address(),
        POLY_SIGNATURE: signature,
        POLY_TIMESTAMP: str(timestamp),
        POLY_NONCE: str(n),
    }

    return headers


def create_level_2_headers(signer: Signer, creds: ApiCreds, request_args: RequestArgs):
    """Creates Level 2 Poly headers for a request."""
    timestamp = str(int(datetime.now(tz=UTC).timestamp()))

    hmac_sig = build_hmac_signature(
        creds.secret,
        timestamp,
        request_args.method,
        request_args.request_path,
        request_args.body,
    )

    return {
        POLY_ADDRESS: signer.address(),
        POLY_SIGNATURE: hmac_sig,
        POLY_TIMESTAMP: timestamp,
        POLY_API_KEY: creds.key,
        POLY_PASSPHRASE: creds.passphrase,
    }



import base64
import hashlib
import hmac


def build_hmac_signature(
    secret: str, timestamp: str, method: str, request_path: str, body=None,
):
    """Creates an HMAC signature by signing a payload with the secret."""
    base64_secret = base64.urlsafe_b64decode(secret)
    message = str(timestamp) + str(method) + str(request_path)
    if body:
        # NOTE: Necessary to replace single quotes with double quotes
        # to generate the same hmac message as go and typescript
        message += str(body).replace("'", '"')

    h = hmac.new(base64_secret, bytes(message, "utf-8"), hashlib.sha256)

    # ensure base64 encoded
    return (base64.urlsafe_b64encode(h.digest())).decode("utf-8")



from py_order_utils.builders import OrderBuilder as UtilsOrderBuilder
from py_order_utils.model import (
    BUY as UTILS_BUY,
)
from py_order_utils.model import (
    EOA,
    OrderData,
    SignedOrder,
)
from py_order_utils.model import (
    SELL as UTILS_SELL,
)
from py_order_utils.signer import Signer as UtilsSigner

from ...types.clob_types import (
    CreateOrderOptions,
    MarketOrderArgs,
    OrderArgs,
    OrderSummary,
    OrderType,
    RoundConfig,
    TickSize,
)
from ..config import get_contract_config
from ..constants import BUY, SELL
from ..exceptions import LiquidityError
from ..signing.signer import Signer
from .helpers import (
    decimal_places,
    round_down,
    round_normal,
    round_up,
    to_token_decimals,
)

ROUNDING_CONFIG: dict[TickSize, RoundConfig] = {
    "0.1": RoundConfig(price=1, size=2, amount=3),
    "0.01": RoundConfig(price=2, size=2, amount=4),
    "0.001": RoundConfig(price=3, size=2, amount=5),
    "0.0001": RoundConfig(price=4, size=2, amount=6),
}


class OrderBuilder:
    def __init__(self, signer: Signer, sig_type=None, funder=None):
        self.signer = signer

        # Signature type used sign orders, defaults to EOA type
        self.sig_type = sig_type if sig_type is not None else EOA

        # Address which holds funds to be used.
        # Used for Polymarket proxy wallets and other smart contract wallets
        # Defaults to the address of the signer
        self.funder = funder if funder is not None else self.signer.address()

    def get_order_amounts(
            self, side: str, size: float, price: float, round_config: RoundConfig,
    ):
        raw_price = round_normal(price, round_config.price)

        if side == BUY:
            raw_taker_amt = round_down(size, round_config.size)

            raw_maker_amt = raw_taker_amt * raw_price
            if decimal_places(raw_maker_amt) > round_config.amount:
                raw_maker_amt = round_up(raw_maker_amt, round_config.amount + 4)
                if decimal_places(raw_maker_amt) > round_config.amount:
                    raw_maker_amt = round_down(raw_maker_amt, round_config.amount)

            maker_amount = to_token_decimals(raw_maker_amt)
            taker_amount = to_token_decimals(raw_taker_amt)

            return UTILS_BUY, maker_amount, taker_amount
        if side == SELL:
            raw_maker_amt = round_down(size, round_config.size)

            raw_taker_amt = raw_maker_amt * raw_price
            if decimal_places(raw_taker_amt) > round_config.amount:
                raw_taker_amt = round_up(raw_taker_amt, round_config.amount + 4)
                if decimal_places(raw_taker_amt) > round_config.amount:
                    raw_taker_amt = round_down(raw_taker_amt, round_config.amount)

            maker_amount = to_token_decimals(raw_maker_amt)
            taker_amount = to_token_decimals(raw_taker_amt)

            return UTILS_SELL, maker_amount, taker_amount
        msg = f"order_args.side must be '{BUY}' or '{SELL}'"
        raise ValueError(msg)

    def get_market_order_amounts(
            self, side: str, amount: float, price: float, round_config: RoundConfig,
    ):
        raw_price = round_normal(price, round_config.price)

        if side == BUY:
            raw_maker_amt = round_down(amount, round_config.size)
            raw_taker_amt = raw_maker_amt / raw_price
            if decimal_places(raw_taker_amt) > round_config.amount:
                raw_taker_amt = round_up(raw_taker_amt, round_config.amount + 4)
                if decimal_places(raw_taker_amt) > round_config.amount:
                    raw_taker_amt = round_down(raw_taker_amt, round_config.amount)

            maker_amount = to_token_decimals(raw_maker_amt)
            taker_amount = to_token_decimals(raw_taker_amt)

            return UTILS_BUY, maker_amount, taker_amount

        if side == SELL:
            raw_maker_amt = round_down(amount, round_config.size)

            raw_taker_amt = raw_maker_amt * raw_price
            if decimal_places(raw_taker_amt) > round_config.amount:
                raw_taker_amt = round_up(raw_taker_amt, round_config.amount + 4)
                if decimal_places(raw_taker_amt) > round_config.amount:
                    raw_taker_amt = round_down(raw_taker_amt, round_config.amount)

            maker_amount = to_token_decimals(raw_maker_amt)
            taker_amount = to_token_decimals(raw_taker_amt)

            return UTILS_SELL, maker_amount, taker_amount
        msg = f"order_args.side must be '{BUY}' or '{SELL}'"
        raise ValueError(msg)

    def create_order(
            self, order_args: OrderArgs, options: CreateOrderOptions,
    ) -> SignedOrder:
        """Creates and signs an order."""
        side, maker_amount, taker_amount = self.get_order_amounts(
            order_args.side,
            order_args.size,
            order_args.price,
            ROUNDING_CONFIG[options.tick_size],
        )

        data = OrderData(
            maker=self.funder,
            taker=order_args.taker,
            tokenId=order_args.token_id,
            makerAmount=str(maker_amount),
            takerAmount=str(taker_amount),
            side=side,
            feeRateBps=str(order_args.fee_rate_bps),
            nonce=str(order_args.nonce),
            signer=self.signer.address(),
            expiration=str(order_args.expiration),
            signatureType=self.sig_type,
        )

        contract_config = get_contract_config(
            self.signer.get_chain_id(), options.neg_risk,
        )

        order_builder = UtilsOrderBuilder(
            contract_config.exchange,
            self.signer.get_chain_id(),
            UtilsSigner(key=self.signer.private_key),
        )

        return order_builder.build_signed_order(data)

    def create_market_order(
            self, order_args: MarketOrderArgs, options: CreateOrderOptions,
    ) -> SignedOrder:
        """Creates and signs a market order."""
        side, maker_amount, taker_amount = self.get_market_order_amounts(
            order_args.side,
            order_args.amount,
            order_args.price,
            ROUNDING_CONFIG[options.tick_size],
        )

        data = OrderData(
            maker=self.funder,
            taker=order_args.taker,
            tokenId=order_args.token_id,
            makerAmount=str(maker_amount),
            takerAmount=str(taker_amount),
            side=side,
            feeRateBps=str(order_args.fee_rate_bps),
            nonce=str(order_args.nonce),
            signer=self.signer.address(),
            expiration="0",
            signatureType=self.sig_type,
        )

        contract_config = get_contract_config(
            self.signer.get_chain_id(), options.neg_risk,
        )

        order_builder = UtilsOrderBuilder(
            contract_config.exchange,
            self.signer.get_chain_id(),
            UtilsSigner(key=self.signer.private_key),
        )

        return order_builder.build_signed_order(data)

    def calculate_buy_market_price(
            self,
            asks: list[OrderSummary], # expected to be sorted from worst to best price (high to low)
            amount_to_match: float, # in usdc
            order_type: OrderType,
    ) -> float:
        if not asks:
            msg = "No ask orders available"
            raise LiquidityError(msg)

        sum = 0
        for p in reversed(asks):
            sum += float(p.size) * float(p.price)
            if sum >= amount_to_match:
                return float(p.price)

        if order_type == OrderType.FOK:
            msg = "no match"
            raise ValueError(msg)

        return float(asks[0].price)

    def calculate_sell_market_price(
            self,
            bids: list[OrderSummary], # expected to be sorted from worst to best price (low to high)
            amount_to_match: float, # in shares
            order_type: OrderType,
    ) -> float:
        if not bids:
            msg = "No bid orders available"
            raise LiquidityError(msg)

        sum = 0
        for p in reversed(bids):
            sum += float(p.size)
            if sum >= amount_to_match:
                print(f"market order price {p.price}")
                return float(p.price)


        if order_type == OrderType.FOK:
            msg = "no match"
            raise ValueError(msg)

        return float(bids[0].price)



import hashlib
from decimal import ROUND_CEILING, ROUND_FLOOR, ROUND_HALF_UP, Decimal

from ...types.clob_types import OrderBookSummary, TickSize


def round_down(x: float, sig_digits: int) -> float:
    exp = Decimal(1).scaleb(-sig_digits)
    return float(Decimal(str(x)).quantize(exp=exp, rounding=ROUND_FLOOR))


def round_normal(x: float, sig_digits: int) -> float:
    exp = Decimal(1).scaleb(-sig_digits)
    return float(Decimal(str(x)).quantize(exp=exp, rounding=ROUND_HALF_UP))


def round_up(x: float, sig_digits: int) -> float:
    exp = Decimal(1).scaleb(-sig_digits)
    return float(Decimal(str(x)).quantize(exp=exp, rounding=ROUND_CEILING))


def to_token_decimals(x: float) -> int:
    exp = Decimal(1)
    return int(
        Decimal(str(x)) * Decimal(10**6).quantize(exp=exp, rounding=ROUND_HALF_UP),
    )


def decimal_places(x: float) -> int:
    """
    Returns the number of decimal places in a numeric value.

    Assumes x is always a finite, non-special value (not NaN or Infinity).
    """
    exponent = Decimal(str(x)).as_tuple().exponent
    if not isinstance(exponent, int):
        msg = "Input must be a finite float."
        raise TypeError(msg)
    return max(0, -exponent)



def generate_orderbook_summary_hash(orderbook: OrderBookSummary) -> str:
    """Compute hash while forcing empty string for hash field."""
    server_hash = orderbook.hash
    orderbook.hash = ""
    computed_hash = hashlib.sha1(
        str(orderbook.model_dump_json(by_alias=True)).encode("utf-8"),
    ).hexdigest()
    orderbook.hash = server_hash
    return computed_hash


def order_to_json(order, owner, order_type) -> dict:
    return {"order": order.dict(), "owner": owner, "orderType": order_type.value}


def is_tick_size_smaller(a: TickSize, b: TickSize) -> bool:
    return float(a) < float(b)


def price_valid(price: float, tick_size: TickSize) -> bool:
    return float(tick_size) <= price <= 1 - float(tick_size)



import logging
from datetime import datetime
from enum import Enum
from typing import Any, Literal, Optional, TypeVar, Union

from py_order_utils.model import SignedOrder
from pydantic import (
    BaseModel,
    ConfigDict,
    Field,
    RootModel,
    ValidationError,
    ValidationInfo,
    ValidatorFunctionWrapHandler,
    field_serializer,
    field_validator,
)

from ..types.common import EthAddress, Keccak256, TimeseriesPoint
from ..utilities.constants import ZERO_ADDRESS

logger = logging.getLogger(__name__)

class ApiCreds(BaseModel):
    key: str = Field(alias="apiKey")
    secret: str
    passphrase: str


class RequestArgs(BaseModel):
    method: Literal["GET", "POST", "DELETE"]
    request_path: str
    body: Any = None


class TokenValue(BaseModel):
    token_id: str
    value: float


class Midpoint(TokenValue):
    pass


class Spread(TokenValue):
    pass


class TokenValueDict(RootModel):
    root: dict[str, float]


class BookParams(BaseModel):
    token_id: str
    side: Literal["BUY", "SELL"]


class Price(BookParams):
    price: float


class BidAsk(BaseModel):
    BUY: Optional[float] = None  # Price buyers are willing to pay
    SELL: Optional[float] = None  # Price sellers are willing to accept


class TokenBidAsk(BidAsk):
    token_id: str


class TokenBidAskDict(RootModel):
    root: dict[str, BidAsk]


class Token(BaseModel):
    token_id: str
    outcome: str
    price: float
    winner: Optional[bool] = None


class PriceHistory(BaseModel):
    token_id: str
    history: list[TimeseriesPoint]


T = TypeVar("T")


class PaginatedResponse[T](BaseModel):
    data: list[T]
    next_cursor: str
    limit: int
    count: int


class RewardRate(BaseModel):
    asset_address: str
    rewards_daily_rate: float


class RewardConfig(BaseModel):
    asset_address: str
    rewards_daily_rate: float = Field(alias="rate_per_day")

    start_date: datetime
    end_date: datetime

    reward_id: Optional[str] = Field(None, alias="id")
    total_rewards: float
    total_days: Optional[int] = None

    @field_validator("reward_id", mode="before")
    def convert_id_to_str(cls, v):
        if isinstance(v, int):
            return str(v)
        return v


class Rewards(BaseModel):
    rates: Optional[list[RewardRate]]
    rewards_min_size: int = Field(alias="min_size")
    rewards_max_spread: float = Field(alias="max_spread")

class EarnedReward(BaseModel):
    asset_address: EthAddress
    earnings: float
    asset_rate: float

class DailyEarnedReward(BaseModel):
    date: datetime
    asset_address: EthAddress
    maker_address: EthAddress
    earnings: float
    asset_rate: float

class PolymarketRewardItem(BaseModel):
    market_id: str
    condition_id: Keccak256
    question: str
    market_slug: str
    market_description: str
    event_slug: str
    image: str
    maker_address: EthAddress
    tokens: list[Token]
    rewards_config: list[RewardConfig]
    earnings: list[EarnedReward]
    rewards_max_spread: float
    rewards_min_size: float
    earning_percentage: float
    spread: float
    market_competitiveness: float


class MarketRewards(BaseModel):
    condition_id: Keccak256
    question: str
    market_slug: str
    event_slug: str
    image: str
    tokens: list[Token]
    rewards_config: list[RewardConfig]
    rewards_max_spread: float
    rewards_min_size: int
    market_competitiveness: float

class ClobMarket(BaseModel):
    # Core market information
    token_ids: list[Token] = Field(alias="tokens")
    condition_id: Keccak256
    question_id: Keccak256
    question: str
    description: str
    market_slug: str
    end_date_iso: Optional[datetime]
    game_start_time: Optional[datetime] = None
    seconds_delay: int

    # Order book settings
    enable_order_book: bool
    accepting_orders: bool
    accepting_order_timestamp: Optional[datetime]
    minimum_order_size: float
    minimum_tick_size: float

    # Status flags
    active: bool
    closed: bool
    archived: bool

    # Negative risk settings
    neg_risk: bool
    neg_risk_market_id: Keccak256
    neg_risk_request_id: Keccak256

    # Fee structure
    fpmm: str
    maker_base_fee: float
    taker_base_fee: float

    # Features
    notifications_enabled: bool
    is_50_50_outcome: bool

    # Visual representation
    icon: str
    image: str

    # Rewards configuration
    rewards: Optional[Rewards]

    # Tags
    tags: Optional[list[str]]

    @field_validator("neg_risk_market_id", "neg_risk_request_id", mode="wrap")
    @classmethod
    def validate_neg_risk_fields(cls, value: str, handler: ValidatorFunctionWrapHandler, info: ValidationInfo) -> Optional[str]:
        try:
            return handler(value)
        except ValidationError as e:
            neg_risk = info.data.get("neg_risk", False)
            active = info.data.get("active", False)
            if not neg_risk and value == "":
                return value
            if not active:
                return value
            if neg_risk and value == "":
                for _ in e.errors():
                    msg = ("Poorly setup market: negative risk is True, but either neg_risk_market_id or neg_risk_request_id is missing. "
                           f" Question: {info.data.get("question")}; Market slug: {info.data.get('market_slug')} \n")
                    logger.warning(msg)
    @field_validator("condition_id", "question_id", mode="wrap")
    @classmethod
    def validate_condition_fields(cls, value: str, handler: ValidatorFunctionWrapHandler, info: ValidationInfo) -> str:
        try:
            return handler(value)
        except ValueError:
            active = info.data.get("active", False)
            if not active:
                return value
            raise

class OpenOrder(BaseModel):
    order_id: Keccak256 = Field(alias="id")
    status: str
    owner: str
    maker_address: str
    condition_id: str = Field(alias="market")
    token_id: str = Field(alias="asset_id")
    side: Literal["BUY", "SELL"]
    original_size: float
    size_matched: float
    price: float
    outcome: str
    expiration: datetime
    order_type: Literal["GTC", "GTD"]
    associate_trades: list[str]
    created_at: datetime

class MakerOrder(BaseModel):
    token_id: str = Field(alias="asset_id")
    order_id: Keccak256
    maker_address: EthAddress
    owner: str
    matched_amount: float
    price: float
    outcome: str
    fee_rate_bps: float

class PolygonTrade(BaseModel):
    trade_id: str = Field(alias="id")
    taker_order_id: Keccak256
    condition_id: Keccak256 = Field(alias="market")
    id: str
    side: Literal["BUY", "SELL"]
    size: float
    fee_rate_bps: float
    price: float
    status: str # change to literals MINED, CONFIRMED
    match_time: datetime
    last_update: datetime
    outcome: str
    bucket_index: int
    owner: str
    maker_address: EthAddress
    transaction_hash: Keccak256
    maker_orders: list[MakerOrder]
    trader_side: Literal["TAKER", "MAKER"]

class TradeParams(BaseModel):
    id: Optional[str] = None
    maker_address: Optional[str] = None
    market: Optional[str] = None
    asset_id: Optional[str] = None
    before: Optional[int] = None
    after: Optional[int] = None


class OpenOrderParams(BaseModel):
    order_id: Optional[str] = None
    condition_id: Optional[str] = None
    token_id: Optional[str] = None


class DropNotificationParams(BaseModel):
    ids: Optional[list[str]]= None


class OrderSummary(BaseModel):
    price: Optional[float] = None
    size: Optional[float] = None

class PriceLevel(OrderSummary):
    side: Literal["BUY", "SELL"]


class OrderBookSummary(BaseModel):
    condition_id: Optional[Keccak256] = Field(None, alias="market")
    token_id: Optional[str] = Field(None, alias="asset_id")
    timestamp: Optional[datetime] = None
    hash: Optional[str] = None
    bids: Optional[list[OrderSummary]] = None
    asks: Optional[list[OrderSummary]] = None

    @field_serializer("bids", "asks")
    def serialize_sizes(self, orders: list[OrderSummary]) -> list[dict]:
        return [
            {
                "price": f"{order.price:.3f}".rstrip("0").rstrip("."),
                "size": f"{order.size:.2f}".rstrip("0").rstrip("."),
            }
            for order in orders
        ]

    @field_serializer("timestamp")
    def serialize_timestamp(self, ts: datetime) -> str:
        # Convert to millisecond timestamp string without decimal places
        return str(int(ts.timestamp() * 1000))


class AssetType(str, Enum):
    COLLATERAL = "COLLATERAL"
    CONDITIONAL = "CONDITIONAL"


class BalanceAllowanceParams(BaseModel):
    asset_type: Optional[AssetType] = None
    token_id: Optional[str] = None
    signature_type: int = -1


class OrderType(str, Enum):
    GTC = "GTC"  # Good Till Cancelled
    GTD = "GTD"  # Good Till Date
    FOK = "FOK"  # Fill or Kill
    FAK = "FAK"  # Fill and Kill


TickSize = Literal["0.1", "0.01", "0.001", "0.0001"]


class CreateOrderOptions(BaseModel):
    tick_size: TickSize
    neg_risk: bool


class PartialCreateOrderOptions(BaseModel):
    tick_size: Optional[TickSize] = None
    neg_risk: Optional[bool] = None


class RoundConfig(BaseModel):
    price: int
    size: int
    amount: int

class OrderArgs(BaseModel):
    token_id: str
    """
    TokenID of the Conditional token asset being traded
    """

    price: float
    """
    Price used to create the order
    """

    size: float
    """
    Size in terms of the ConditionalToken
    """

    side: str
    """
    Side of the order
    """

    fee_rate_bps: int = 0
    """
    Fee rate, in basis points, charged to the order maker, charged on proceeds
    """

    nonce: int = 0
    """
    Nonce used for onchain cancellations
    """

    expiration: int = 0
    """
    Timestamp after which the order is expired.
    """

    taker: str = ZERO_ADDRESS
    """
    Address of the order taker. The zero address is used to indicate a public order.
    """


class MarketOrderArgs(BaseModel):
    token_id: str
    """
    TokenID of the Conditional token asset being traded
    """

    amount: float
    """
    BUY orders: $$$ Amount to buy
    SELL orders: Shares to sell
    """

    side: str
    """
    Side of the order
    """

    price: float = 0
    """
    Price used to create the order
    """

    fee_rate_bps: int = 0
    """
    Fee rate, in basis points, charged to the order maker, charged on proceeds.
    """

    nonce: int = 0
    """
    Nonce used for onchain cancellations.
    """

    taker: str = ZERO_ADDRESS
    """
    Address of the order taker. The zero address is used to indicate a public order.
    """

    order_type: OrderType = OrderType.FOK

class PostOrdersArgs(BaseModel):
    order: SignedOrder
    order_type: OrderType = OrderType.GTC

    model_config = ConfigDict(arbitrary_types_allowed=True)

class ContractConfig(BaseModel):
    """Contract Configuration."""

    exchange: EthAddress
    """
    The exchange contract responsible for matching orders.
    """

    collateral: EthAddress
    """
    The ERC20 token used as collateral for the exchange's markets.
    """

    conditional_tokens: EthAddress
    """
    The ERC1155 conditional tokens contract.
    """


class OrderPostResponse(BaseModel):
    error_msg: str = Field(alias="errorMsg")
    order_id: Union[Keccak256, Literal[""]] = Field(alias="orderID")
    taking_amount: str = Field(alias="takingAmount")
    making_amount: str = Field(alias="makingAmount")
    status: str = Literal["live", "matched", "delayed"]
    success: bool

class OrderCancelResponse(BaseModel):
    not_canceled: Optional[dict[Keccak256, str]]
    canceled: Optional[list[Keccak256]]



from typing import Literal

from gql import Client, gql
from gql.transport.httpx import HTTPXAsyncTransport, HTTPXTransport

from ..utilities.config import GRAPHQL_ENDPOINTS

EndpointName = Literal[
            "activity_subgraph",
            "fpmm_subgraph",
            "open_interest_subgraph",
            "orderbook_subgraph",
            "pnl_subgraph",
            "positions_subgraph",
            "sports_oracle_subgraph",
            "wallet_subgraph",
            ]


class PolymarketGraphQLClient:
    """Synchronous GraphQL client for Polymarket subgraphs."""

    def __init__(self, endpoint_name: EndpointName) -> None:
        endpoint_url = GRAPHQL_ENDPOINTS[endpoint_name]
        self.transport = HTTPXTransport(url=endpoint_url)
        self.client = Client(transport=self.transport, fetch_schema_from_transport=False)

    def query(self, query_string: str) -> dict:
        with self.client as session:
            return session.execute(gql(query_string))


class AsyncPolymarketGraphQLClient:
    """Asynchronous GraphQL client for Polymarket subgraphs."""

    def __init__(self, endpoint_name: EndpointName) -> None:
        endpoint_url = GRAPHQL_ENDPOINTS[endpoint_name]
        self.transport = HTTPXAsyncTransport(url=endpoint_url)
        self.client = Client(transport=self.transport, fetch_schema_from_transport=False)

    async def query(self, query_string: str) -> dict:
        async with self.client as session:
            return await session.execute(gql(query_string))



from datetime import datetime
from typing import Literal, Optional

from pydantic import AliasChoices, BaseModel, Field, field_validator

from ..types.clob_types import MakerOrder, OrderBookSummary, TickSize
from ..types.common import EthAddress, Keccak256, TimeseriesPoint

# wss://ws-subscriptions-clob.polymarket.com/ws/market types

class PriceChange(BaseModel):
    best_ask: float = Field(validation_alias=AliasChoices("ba", "best_ask"))
    best_bid: float = Field(validation_alias=AliasChoices("bb", "best_bid"))
    price: float = Field(validation_alias=AliasChoices("p", "price"))
    size: float = Field(validation_alias=AliasChoices("s", "size"))
    side: Literal["BUY", "SELL"] = Field(validation_alias=AliasChoices("si", "side"))
    token_id: str = Field(validation_alias=AliasChoices("a", "asset_id"))
    hash: str = Field(validation_alias=AliasChoices("h", "hash"))

class PriceChanges(BaseModel):
    condition_id: Keccak256 = Field(validation_alias=AliasChoices("m", "market"))
    price_changes: list[PriceChange] = Field(validation_alias=AliasChoices("pc", "price_changes"))
    timestamp: datetime = Field(validation_alias=AliasChoices("t", "timestamp"))

class TickSizeChange(BaseModel):
    token_id: str = Field(alias="asset_id")
    condition_id: Keccak256 = Field(alias="market")
    old_tick_size: TickSize
    new_tick_size: TickSize

class LastTradePrice(BaseModel):
    price: float
    size: float
    side: Literal["BUY", "SELL"]
    token_id: str = Field(alias="asset_id")
    condition_id: Keccak256 = Field(alias="market")
    fee_rate_bps: float

class OrderBookSummaryEvent(OrderBookSummary):
    event_type: Literal["book"]

class PriceChangeEvent(PriceChanges):
    event_type: Literal["price_change"]

class TickSizeChangeEvent(TickSizeChange):
    side: Literal["BUY", "SELL"]
    timestamp: datetime
    event_type: Literal["tick_size_change"]

class LastTradePriceEvent(LastTradePrice):
    timestamp: datetime
    event_type: Literal["last_trade_price"]

# wss://ws-subscriptions-clob.polymarket.com/ws/user types

class OrderEvent(BaseModel):
    token_id: str = Field(alias="asset_id")
    condition_id: Keccak256 = Field(alias="market")
    order_id: Keccak256 = Field(alias="id")
    associated_trades: Optional[list[str]] = None # list of trade ids which
    maker_address: EthAddress
    order_owner: str = Field(alias="owner") # api key of order owner
    event_owner: Optional[str] = Field(None, alias="owner") # api key of event owner

    price: float
    side: Literal["BUY", "SELL"]
    size_matched: float
    original_size: float
    outcome: str
    order_type: Literal["GTC", "GTD", "FOK", "FAK"]

    created_at: datetime
    expiration: Optional[datetime] = None
    timestamp: Optional[datetime] = None # time of event

    event_type: Optional[Literal["order"]] = None
    type: Literal["PLACEMENT", "UPDATE", "CANCELLATION"]

    status: Literal["LIVE", "CANCELED", "MATCHED"]

    @field_validator("expiration", mode="before")
    def validate_expiration(cls, v):
        if v == "0":
            return None
        return v

class TradeEvent(BaseModel):
    token_id: str = Field(alias="asset_id")
    condition_id: Keccak256 = Field(alias="market")
    taker_order_id: Keccak256
    maker_orders: list[MakerOrder]
    trade_id: str = Field(alias="id")
    trade_owner: Optional[str] = Field(None, alias="owner") # api key of trade owner
    event_owner: str = Field(alias="owner") # api key of event owner

    price: float
    size: float
    side: Literal["BUY", "SELL"]
    outcome: str

    last_update: datetime # time of last update to trade
    matchtime: Optional[datetime] = None # time trade was matched
    timestamp: Optional[datetime] = None # time of event

    event_type: Optional[Literal["trade"]] = None
    type: Optional[Literal["TRADE"]] = None

    status: Literal["MATCHED", "MINED", "CONFIRMED", "RETRYING", "FAILED"]

# wss://ws-live-data.polymarket.com types

# Payload models
class ActivityTrade(BaseModel):
    token_id: str = Field(alias="asset") # ERC1155 token ID of conditional token being traded
    condition_id: str = Field(alias="conditionId")  # Id of market which is also the CTF condition ID
    event_slug: str = Field(alias="eventSlug")  # Slug of the event
    outcome: str  # Human readable outcome of the market
    outcome_index: int = Field(alias="outcomeIndex")  # Index of the outcome
    price: float  # Price of the trade
    side: Literal["BUY", "SELL"]  # Side of the trade
    size: float  # Size of the trade
    slug: str  # Slug of the market
    timestamp: datetime  # Timestamp of the trade
    title: str  # Title of the event
    transaction_hash: str = Field(alias="transactionHash")  # Hash of the transaction
    proxy_wallet: str = Field(alias="proxyWallet")  # Address of the user proxy wallet
    icon: str  # URL to the market icon image
    name: str  # Name of the user of the trade
    bio: str  # Bio of the user of the trade
    pseudonym: str  # Pseudonym of the user
    profile_image: str = Field(alias="profileImage")  # URL to the user profile image
    profile_image_optimized: Optional[str] = Field(None, alias="profileImageOptimized")

class Comment(BaseModel):
    id: str  # Unique identifier of comment
    body: str  # Content of the comment
    parent_entity_type: str = Field(alias="parentEntityType")  # Type of the parent entity (Event or Series)
    parent_entity_id: int = Field(alias="parentEntityID")  # ID of the parent entity
    parent_comment_id: Optional[str] = Field(None, alias="parentCommentID")  # ID of the parent comment
    user_address: str = Field(alias="userAddress")  # Address of the user
    reply_address: Optional[str] = Field(None, alias="replyAddress")  # Address of the reply user
    created_at: datetime = Field(alias="createdAt")  # Creation timestamp
    updated_at: Optional[datetime] = Field(None, alias="updatedAt")  # Last update timestamp

class Reaction(BaseModel):
    id: str  # Unique identifier of reaction
    comment_id: int = Field(alias="commentID")  # ID of the comment
    reaction_type: str = Field(alias="reactionType")  # Type of the reaction
    icon: Optional[str] = None  # Icon representing the reaction
    user_address: str = Field(alias="userAddress")  # Address of the user
    created_at: datetime = Field(alias="createdAt")  # Creation timestamp

class Request(BaseModel):
    request_id: str = Field(alias="requestId")  # Unique identifier for the request
    proxy_address: str = Field(alias="proxyAddress")  # Proxy address
    user_address: str = Field(alias="userAddress")  # User address
    condition_id: Keccak256 = Field(alias="market")  # Id of market which is also the CTF condition ID
    token_id: str = Field(alias="token")  # ERC1155 token ID of conditional token being traded
    complement_token_id: str = Field(alias="complement")  # Complement ERC1155 token ID of conditional token being traded
    state: Literal["STATE_REQUEST_EXPIRED", "STATE_USER_CANCELED", "STATE_REQUEST_CANCELED", "STATE_MAKER_CANCELED", "STATE_ACCEPTING_QUOTES", "STATE_REQUEST_QUOTED", "STATE_QUOTE_IMPROVED"]  # Current state of the request
    side: Literal["BUY", "SELL"]  # Indicates buy or sell side
    price: float  # Price from in/out sizes
    size_in: float = Field(alias="sizeIn")  # Input size of the request
    size_out: float = Field(alias="sizeOut")  # Output size of the request
    expiry: Optional[datetime] = None

class Quote(BaseModel):
    quote_id: str = Field(alias="quoteId")  # Unique identifier for the quote
    request_id: str = Field(alias="requestId")  # Associated request identifier
    proxy_address: str = Field(alias="proxyAddress")  # Proxy address
    user_address: str = Field(alias="userAddress")  # User address
    condition_id: Keccak256 = Field(alias="condition")  # Id of market which is also the CTF condition ID
    token_id: str = Field(alias="token")  # ERC1155 token ID of conditional token being traded
    complement_token_id: str = Field(alias="complement")  # Complement ERC1155 token ID of conditional token being traded
    state: Literal["STATE_REQUEST_EXPIRED", "STATE_USER_CANCELED", "STATE_REQUEST_CANCELED", "STATE_MAKER_CANCELED", "STATE_ACCEPTING_QUOTES", "STATE_REQUEST_QUOTED", "STATE_QUOTE_IMPROVED"]  # Current state of the quote
    side: Literal["BUY", "SELL"]  # Indicates buy or sell side
    size_in: float = Field(alias="sizeIn")  # Input size of the quote
    size_out: float = Field(alias="sizeOut")  # Output size of the quote
    expiry: Optional[datetime] = None

class CryptoPriceSubscribe(BaseModel):
    data: list[TimeseriesPoint]
    symbol: str

class CryptoPriceUpdate(TimeseriesPoint):
    symbol: str
    full_accuracy_value: str

class AggOrderBookSummary(OrderBookSummary):
    min_order_size: float
    tick_size: TickSize
    neg_risk: bool

class LiveDataClobMarket(BaseModel):
    token_ids: list[str] = Field(alias="asset_ids")
    condition_id: Keccak256 = Field(alias="market")
    min_order_size: float
    tick_size: TickSize
    neg_risk: bool

# Event models
class ActivityTradeEvent(BaseModel):
    payload: ActivityTrade
    timestamp: datetime
    type: Literal["trades"]
    topic: Literal["activity"]

class ActivityOrderMatchEvent(BaseModel):
    payload: ActivityTrade
    timestamp: datetime
    type: Literal["orders_matched"]
    topic: Literal["activity"]

class CommentEvent(BaseModel):
    payload: Comment
    timestamp: datetime
    type: Literal["comment_created", "comment_removed"]
    topic: Literal["comments"]

class ReactionEvent(BaseModel):
    payload: Reaction
    timestamp: datetime
    type: Literal["reaction_created", "reaction_removed"]
    topic: Literal["comments"]

class RequestEvent(BaseModel):
    payload: Request
    timestamp: datetime
    type: Literal["request_created", "request_edited", "request_canceled", "request_expired"]
    topic: Literal["rfq"]

class QuoteEvent(BaseModel):
    payload: Quote
    timestamp: datetime
    type: Literal["quote_created", "quote_edited", "quote_canceled", "quote_expired"]
    topic: Literal["rfq"]

class CryptoPriceUpdateEvent(BaseModel):
    payload: CryptoPriceUpdate
    timestamp: datetime
    connection_id: str
    type: Literal["update"]
    topic: Literal["crypto_prices", "crypto_prices_chainlink"]

class CryptoPriceSubscribeEvent(BaseModel):
    payload: CryptoPriceSubscribe
    timestamp: datetime
    type: Literal["subscribe"]
    topic: Literal["crypto_prices", "crypto_prices_chainlink"]

class LiveDataOrderBookSummaryEvent(BaseModel):
    payload: list[AggOrderBookSummary] | AggOrderBookSummary
    timestamp: datetime
    connection_id: str
    type: Literal["agg_orderbook"]
    topic: Literal["clob_market"]

class LiveDataPriceChangeEvent(BaseModel):
    payload: PriceChanges
    timestamp: datetime
    connection_id: str
    type: Literal["price_change"]
    topic: Literal["clob_market"]

class LiveDataLastTradePriceEvent(BaseModel):
    payload: LastTradePrice
    timestamp: datetime
    connection_id: str
    type: Literal["last_trade_price"]
    topic: Literal["clob_market"]

class LiveDataTickSizeChangeEvent(BaseModel):
    payload: TickSizeChange
    timestamp: datetime
    connection_id: str
    type: Literal["tick_size_change"]
    topic: Literal["clob_market"]

class MarketStatusChangeEvent(BaseModel):
    payload: LiveDataClobMarket
    timestamp: datetime
    connection_id: str
    type: Literal["market_created", "market_resolved"]
    topic: Literal["clob_market"]

class LiveDataOrderEvent(BaseModel):
    payload: OrderEvent
    timestamp: datetime
    connection_id: str
    type: Literal["order"]
    topic: Literal["clob_user"]

class LiveDataTradeEvent(BaseModel):
    payload: TradeEvent
    timestamp: datetime
    connection_id: str
    type: Literal["trade"]
    topic: Literal["clob_user"]

class ErrorEvent(BaseModel):
    message: str
    connection_id: str = Field(alias="connectionId")
    request_id: str = Field(alias="requestId")





from eth_utils import keccak
from poly_eip712_structs import make_domain
from py_order_utils.utils import prepend_zx

from ..signing.model import ClobAuth
from ..signing.signer import Signer

CLOB_DOMAIN_NAME = "ClobAuthDomain"
CLOB_VERSION = "1"
MSG_TO_SIGN = "This message attests that I control the given wallet"


def get_clob_auth_domain(chain_id: int):
    return make_domain(name=CLOB_DOMAIN_NAME, version=CLOB_VERSION, chainId=chain_id)


def sign_clob_auth_message(signer: Signer, timestamp: int, nonce: int) -> str:
    clob_auth_msg = ClobAuth(
        address=signer.address(),
        timestamp=str(timestamp),
        nonce=nonce,
        message=MSG_TO_SIGN,
    )
    chain_id = signer.get_chain_id()
    auth_struct_hash = prepend_zx(
        keccak(clob_auth_msg.signable_bytes(get_clob_auth_domain(chain_id))).hex(),
    )
    return prepend_zx(signer.sign(auth_struct_hash))



class InvalidPriceError(Exception):
    pass

class InvalidTickSizeError(Exception):
    pass

class InvalidFeeRateError(Exception):
    pass

class LiquidityError(Exception):
    pass

class MissingOrderbookError(Exception):
    pass

class AuthenticationRequiredError(ValueError):
    """Raised when authentication credentials are required but not provided."""



import json
import random
import string
from datetime import datetime
from typing import Literal, Optional, Union
from urllib.parse import urljoin

import httpx

from ..types.gamma_types import Event, EventList, GammaMarket


def generate_random_id(length=16):
    characters = string.ascii_letters + string.digits
    random_id = "".join(random.choices(characters, k=length))
    return random_id


class PolymarketGammaClient:
    def __init__(self, base_url: str = "https://gamma-api.polymarket.com"):
        self.base_url = base_url
        self.client = httpx.Client(http2=True, timeout=30.0)

    def _build_url(self, endpoint: str) -> str:
        return urljoin(self.base_url, endpoint)

    def get_markets(
            self,
            limit: Optional[int] = None,
            offset: Optional[int] = None,
            order: Optional[str] = None,
            ascending: bool = True,
            archived: Optional[bool] = None,
            active: Optional[bool] = None,
            closed: Optional[bool] = None,
            slugs: Optional[list[str]] = None,
            market_ids: Optional[list[int]] = None,
            token_ids: Optional[list[str]] = None,
            condition_ids: Optional[list[str]] = None,
            tag_id: Optional[int] = None,
            related_tags: Optional[bool] = False,
            liquidity_num_min: Optional[float] = None,
            liquidity_num_max: Optional[float] = None,
            volume_num_min: Optional[float] = None,
            volume_num_max: Optional[float] = None,
            start_date_min: Optional[datetime] = None,
            start_date_max: Optional[datetime] = None,
            end_date_min: Optional[datetime] = None,
            end_date_max: Optional[datetime] = None,
    ) -> list[GammaMarket]:
        params = {}
        if limit:
            params["limit"] = limit
        if offset:
            params["offset"] = offset
        if order:
            params["order"] = order
            params["ascending"] = ascending
        if slugs:
            params["slug"] = slugs
        if archived is not None:
            params["archived"] = archived
        if active is not None:
            params["active"] = active
        if closed is not None:
            params["closed"] = closed
        if market_ids:
            params["id"] = market_ids
        if token_ids:
            params["clob_token_ids"] = token_ids
        if condition_ids:
            params["condition_ids"] = condition_ids
        if liquidity_num_min:
            params["liquidity_num_min"] = liquidity_num_min
        if liquidity_num_max:
            params["liquidity_num_max"] = liquidity_num_max
        if volume_num_min:
            params["volume_num_min"] = volume_num_min
        if volume_num_max:
            params["volume_num_max"] = volume_num_max
        if start_date_min:
            params["start_date_min"] = start_date_min.isoformat()
        if start_date_max:
            params["start_date_max"] = start_date_max.isoformat()
        if end_date_min:
            params["end_date_min"] = end_date_min.isoformat()
        if end_date_max:
            params["end_date_max"] = end_date_max.isoformat()
        if tag_id:
            params["tag_id"] = tag_id
            if related_tags:
                params["related_tags"] = related_tags

        response = self.client.get(self._build_url("/markets"), params=params)
        response.raise_for_status()
        return [GammaMarket(**market) for market in response.json()]

    def get_market(self, market_id: str) -> GammaMarket:
        """Get a GammaMarket by market_id."""
        response = self.client.get(self._build_url(f"/markets/{market_id}"))
        response.raise_for_status()
        return GammaMarket(**response.json())

    def get_events(
            self,
            limit: Optional[int] = None,
            offset: Optional[int] = None,
            order: Optional[str] = None,
            ascending: bool = True,
            event_ids: Optional[Union[str, list[str]]] = None,
            slugs: Optional[list[str]] = None,
            archived: Optional[bool] = None,
            active: Optional[bool] = None,
            closed: Optional[bool] = None,
            liquidity_min: Optional[float] = None,
            liquidity_max: Optional[float] = None,
            volume_min: Optional[float] = None,
            volume_max: Optional[float] = None,
            start_date_min: Optional[datetime] = None,
            start_date_max: Optional[datetime] = None,
            end_date_min: Optional[datetime] = None,
            end_date_max: Optional[datetime] = None,
            tag: Optional[str] = None,
            tag_id: Optional[int] = None,
            tag_slug: Optional[str] = None,
            related_tags: bool = False,
    ) -> list[Event]:
        params = {}
        if limit:
            params["limit"] = limit
        if offset:
            params["offset"] = offset
        if order:
            params["order"] = order
            params["ascending"] = ascending
        if event_ids:
            params["id"] = event_ids
        if slugs:
            params["slug"] = slugs
        if archived is not None:
            params["archived"] = archived
        if active is not None:
            params["active"] = active
        if closed is not None:
            params["closed"] = closed
        if liquidity_min:
            params["liquidity_min"] = liquidity_min
        if liquidity_max:
            params["liquidity_max"] = liquidity_max
        if volume_min:
            params["volume_min"] = volume_min
        if volume_max:
            params["volume_max"] = volume_max
        if start_date_min:
            params["start_date_min"] = start_date_min.isoformat()
        if start_date_max:
            params["start_date_max"] = start_date_max.isoformat()
        if end_date_min:
            params["end_date_min"] = end_date_min.isoformat()
        if end_date_max:
            params["end_date_max"] = end_date_max.isoformat()
        if tag:
            params["tag"] = tag
        elif tag_id:
            params["tag_id"] = tag_id
            if related_tags:
                params["related_tags"] = related_tags
        elif tag_slug:
            params["tag_slug"] = tag_slug

        response = self.client.get(self._build_url("/events"), params=params)
        response.raise_for_status()
        return [Event(**event) for event in response.json()]

    def get_event(self, event_id: int) -> Event:
        response = self.client.get(self._build_url(f"/events/{event_id}"))
        response.raise_for_status()
        return Event(**response.json())

    def get_all_events(
            self,
            order: Optional[str] = None,
            ascending: bool = True,
            event_ids: Optional[Union[str, list[str]]] = None,
            slugs: Optional[list[str]] = None,
            archived: Optional[bool] = None,
            active: Optional[bool] = None,
            closed: Optional[bool] = None,
            liquidity_min: Optional[float] = None,
            liquidity_max: Optional[float] = None,
            volume_min: Optional[float] = None,
            volume_max: Optional[float] = None,
            start_date_min: Optional[datetime] = None,
            start_date_max: Optional[datetime] = None,
            end_date_min: Optional[datetime] = None,
            end_date_max: Optional[datetime] = None,
            tag: Optional[str] = None,
            tag_id: Optional[int] = None,
            tag_slug: Optional[str] = None,
            related_tags: bool = False,
    ) -> list[Event]:
        offset = 0
        events = []

        while True:
            part = self.get_events(
                limit=500,
                offset=offset,
                order=order,
                ascending=ascending,
                event_ids=event_ids,
                slugs=slugs,
                archived=archived,
                active=active,
                closed=closed,
                liquidity_min=liquidity_min,
                liquidity_max=liquidity_max,
                volume_min=volume_min,
                volume_max=volume_max,
                start_date_min=start_date_min,
                start_date_max=start_date_max,
                end_date_min=end_date_min,
                end_date_max=end_date_max,
                tag=tag,
                tag_id=tag_id,
                tag_slug=tag_slug,
                related_tags=related_tags,
            )
            events.extend(part)

            if len(part) < 500:
                break

            offset += 500

        return events

    def search_events(
            self,
            query: str,
            active: bool = True,
            status: Optional[Literal["active", "resolved"]] = "active",
            sort: Literal["volume", "volume_24hr", "liquidity", "start_date", "end_date", "competitive"] = "volume_24hr",
            page: int = 1,
            limit_per_type: int = 50, # max is 50
            presets: Optional[Literal["EventsHybrid", "EventsTitle"] | list[Literal["EventsHybrid", "EventsTitle"]]] = None,
    ) -> EventList:
        """Search for events by query. Should emulate the website search function."""
        params = {"q": query,"page": page, "limit_per_type": limit_per_type, "events_status": status, "active": active}
        if sort:
            params["sort"] = sort
        if sort == "end_date":
            params["ascending"] = "true"
        if presets:
            params["presets"] = presets
        response = self.client.get(self._build_url("/public-search"), params=params)
        response.raise_for_status()
        return EventList(**response.json())

    def grok_event_summary(self, event_slug: str):
        json_payload = {
            "id": generate_random_id(),
            "messages": [{"role": "user", "content": "", "parts": []}],
        }

        params = {
            "prompt": event_slug,
        }

        with self.client.stream(
            method="POST",
            url="https://polymarket.com/api/grok/event-summary",
            params=params,
            json=json_payload,
        ) as stream:
            messages = []
            citations = []
            seen_urls = set()

            for line_bytes in stream.iter_lines():
                line = (
                    line_bytes.decode() if isinstance(line_bytes, bytes) else line_bytes
                )
                if line.startswith("__SOURCES__:"):
                    sources_json_str = line[len("__SOURCES__:") :]
                    try:
                        sources_obj = json.loads(sources_json_str)
                        for source in sources_obj.get("sources", []):
                            url = source.get("url")
                            if url and url not in seen_urls:
                                citations.append(source)
                                seen_urls.add(url)
                    except json.JSONDecodeError:
                        pass
                else:
                    messages.append(line)
                    print(line, end="")  # or handle message text as needed

        # After reading streamed lines:
        if citations:
            print("\n\nSources:")
            for source in citations:
                print(f"- {source.get('url', 'Unknown URL')}")

    def grok_election_market_explanation(self, candidate_name: str, election_title: str):
        text = f"Provide candidate information for {candidate_name} in the {election_title} on Polymarket."
        json_payload = {
            "id": generate_random_id(),
            "messages": [
                {
                    "role": "user",
                    "content": text,
                    "parts": [{"type": "text", "text": text}],
                },
            ],
        }

        response = self.client.post(
            url="https://polymarket.com/api/grok/election-market-explanation",
            json=json_payload,
        )
        response.raise_for_status()

        parts = [p.strip() for p in response.text.split("**") if p.strip()]
        for i, part in enumerate(parts):
            if ":" in part and i != 0:
                print()
            print(part)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.client.close()



from __future__ import annotations

from datetime import datetime
from typing import Optional

from pydantic import (
    BaseModel,
    Field,
    Json,
    ValidationInfo,
    ValidatorFunctionWrapHandler,
    field_validator,
)

from .common import EthAddress, Keccak256, TimestampWithTZ


class Event(BaseModel):
    # Identifiers
    id: str
    slug: str
    ticker: Optional[str] = None

    # Core event information
    title: str
    description: Optional[str] = None
    resolution_source: Optional[str] = Field(None, alias="resolutionSource")
    category: Optional[str] = None

    # Media URLs
    image: Optional[str] = None
    icon: Optional[str] = None

    # Datetime
    start_date: Optional[datetime] = Field(None, alias="startDate")
    end_date: Optional[datetime] = Field(None, alias="endDate")
    creation_date: Optional[datetime] = Field(None, alias="creationDate")
    created_at: Optional[datetime] = Field(alias="createdAt")
    updated_at: Optional[datetime] = Field(None, alias="updatedAt")
    published_at: Optional[datetime] = Field(None, alias="publishedAt")
    closed_time: Optional[datetime] = Field(None, alias="closedTime")

    # Status flags
    active: bool
    closed: bool
    archived: Optional[bool] = None
    new: Optional[bool] = None
    featured: Optional[bool] = None
    restricted: Optional[bool] = None
    cyom: bool
    automatically_active: Optional[bool] = Field(None, alias="automaticallyActive")

    # Financial metrics
    liquidity: Optional[float] = None
    volume: Optional[float] = None
    open_interest: Optional[int] = Field(None, alias="openInterest")
    competitive: Optional[float] = None
    volume_24hr: Optional[float] = Field(None, alias="volume24hr")
    liquidity_amm: Optional[float] = Field(None, alias="liquidityAmm")
    liquidity_clob: Optional[float] = Field(None, alias="liquidityClob")

    # Related data
    markets: Optional[list[GammaMarket]] = None
    series: Optional[list[Series]] = None
    tags: Optional[list[Tag]] = None

    # User interaction
    comment_count: Optional[int] = Field(None, alias="commentCount")

    # Display and functionality settings
    sort_by: Optional[str] = Field(None, alias="sortBy")
    show_all_outcomes: bool = Field(alias="showAllOutcomes")
    show_market_images: bool = Field(alias="showMarketImages")
    gmp_chart_mode: Optional[str] = Field(None, alias="gmpChartMode")

    # Negative risk settings
    enable_neg_risk: bool = Field(alias="enableNegRisk")
    neg_risk: Optional[bool] = Field(None, alias="negRisk")
    neg_risk_market_id: Optional[str] = Field(None, alias="negRiskMarketID")
    neg_risk_augmented: Optional[bool] = Field(None, alias="negRiskAugmented")

    # Order book settings
    enable_order_book: Optional[bool] = Field(None, alias="enableOrderBook")


class GammaMarket(BaseModel):
    # Identifiers
    id: str
    slug: str
    condition_id: Keccak256 = Field(alias="conditionId")
    question_id: Optional[Keccak256] = Field(None, alias="questionID")

    # Core market information
    question: str
    description: str
    resolution_source: Optional[str] = Field(None, alias="resolutionSource")
    outcome: Optional[list] = None
    outcome_prices: Optional[Json[list[float]] | list[float]] = Field(None, alias="outcomePrices")

    # Media URLs
    image: Optional[str] = None
    icon: Optional[str] = None

    # Datetime
    start_date: Optional[datetime] = Field(None, alias="startDate")
    end_date: Optional[datetime] = Field(None, alias="endDate")
    created_at: datetime = Field(alias="createdAt")
    updated_at: Optional[datetime] = Field(None, alias="updatedAt")
    start_date_iso: Optional[datetime] = Field(None, alias="startDateIso")
    end_date_iso: Optional[datetime] = Field(None, alias="endDateIso")
    deployed_timestamp: Optional[datetime] = Field(None, alias="deployedTimestamp")
    accepting_orders_timestamp: Optional[datetime] = Field(None, alias="acceptingOrdersTimestamp")

    # Status flags
    active: bool
    closed: bool
    archived: bool
    new: Optional[bool] = None
    featured: Optional[bool] = None
    restricted: bool
    ready: bool
    deployed: Optional[bool] = None
    funded: bool
    cyom: bool
    approved: bool

    # Financial metrics
    liquidity: Optional[float] = None
    volume: Optional[float] = None
    volume_num: Optional[float] = Field(None, alias="volumeNum")
    liquidity_num: Optional[float] = Field(None, alias="liquidityNum")
    volume_24hr: Optional[float] = Field(None, alias="volume24hr")
    volume_24hr_clob: Optional[float] = Field(None, alias="volume24hrClob")
    volume_clob: Optional[float] = Field(None, alias="volumeClob")
    liquidity_clob: Optional[float] = Field(None, alias="liquidityClob")
    competitive: Optional[float] = None
    spread: float

    # Order book settings
    enable_order_book: Optional[bool] = Field(None, alias="enableOrderBook")
    order_price_min_tick_size: Optional[float] = Field(
        None, alias="orderPriceMinTickSize",
    )
    order_min_size: Optional[float] = Field(None, alias="orderMinSize")
    accepting_orders: Optional[bool] = Field(None, alias="acceptingOrders")

    # Related data
    events: Optional[list[Event]] = None
    clob_rewards: Optional[list[ClobReward]] = Field(None, alias="clobRewards")

    # User interaction
    comment_count: Optional[int] = Field(None, alias="commentCount")

    # Market maker information
    market_maker_address: str = Field(alias="marketMakerAddress")

    # Additional settings
    group_item_title: Optional[str] = Field(None, alias="groupItemTitle")
    group_item_threshold: Optional[int] = Field(None, alias="groupItemThreshold")
    token_ids: Optional[Json[list[str]] | list[str]] = Field(None, alias="clobTokenIds")
    uma_bond: Optional[int] = Field(None, alias="umaBond")
    uma_reward: Optional[float] = Field(None, alias="umaReward")
    neg_risk: Optional[bool] = Field(None, alias="negRisk")
    pager_duty_notification_enabled: bool = Field(alias="pagerDutyNotificationEnabled")
    review_status: Optional[str] = Field(None, alias="reviewStatus")
    rewards_min_size: int = Field(alias="rewardsMinSize")
    rewards_max_spread: float = Field(alias="rewardsMaxSpread")

    # Resolution information
    submitted_by: Optional[str] = None
    resolved_by: Optional[EthAddress] = Field(None, alias="resolvedBy")
    has_reviewed_dates: Optional[bool] = Field(None, alias="hasReviewedDates")

    @field_validator("condition_id", mode="wrap")
    @classmethod
    def validate_condition_id(
            cls, value: str, handler: ValidatorFunctionWrapHandler, info: ValidationInfo,
    ) -> str:
        try:
            # First attempt standard Keccak256 validation
            return handler(value)
        except ValueError:
            active = info.data.get("active", False)

            # Only allow empty string when inactive
            if not active and value == "":
                return value

            # Re-raise original error for other cases
            raise

class ClobReward(BaseModel):
    # Identifiers
    id: str
    condition_id: Keccak256 = Field(alias="conditionId")

    # Reward information
    asset_address: str = Field(alias="assetAddress")
    rewards_amount: float = Field(alias="rewardsAmount")
    rewards_daily_rate: Optional[float] = Field(None, alias="rewardsDailyRate")

    # Datetime
    start_date: datetime = Field(alias="startDate")
    end_date: datetime = Field(alias="endDate")


class Tag(BaseModel):
    # Identifiers
    id: str
    label: str
    slug: str

    # Display settings
    force_show: Optional[bool] = Field(None, alias="forceShow")
    force_hide: Optional[bool] = Field(None, alias="forceHide")

    # Datetime
    published_at: Optional[TimestampWithTZ | datetime] = Field(None, alias="publishedAt")
    created_at: Optional[datetime] = Field(None, alias="createdAt")
    updated_at: Optional[datetime] = Field(None, alias="updatedAt")

    # User information
    created_by: Optional[int] = Field(None, alias="createdBy")
    updated_by: Optional[int] = Field(None, alias="updatedBy")


class Series(BaseModel):
    # Identifiers
    id: str
    slug: str
    ticker: str
    title: str

    # Series characteristics
    series_type: Optional[str] = Field(None,  alias="seriesType")
    recurrence: Optional[str] = None
    layout: Optional[str] = None

    # Media URLs
    icon: Optional[str] = None
    image: Optional[str] = None

    # Datetime
    start_date: Optional[datetime] = Field(None, alias="startDate")
    created_at: datetime = Field(alias="createdAt")
    updated_at: Optional[datetime] = Field(None, alias="updatedAt")
    published_at: Optional[TimestampWithTZ | datetime] = Field(None, alias="publishedAt")

    # Status flags
    active: Optional[bool] = None
    archived: Optional[bool] = None
    closed: Optional[bool] = None
    featured: Optional[bool] = None
    new: Optional[bool] = None
    restricted: Optional[bool] = None

    # Financial metrics
    liquidity: Optional[float] = None
    volume: Optional[float] = None
    volume_24hr: Optional[float] = Field(None, alias="volume24hr")
    competitive: Optional[str] = None

    # User interaction
    comment_count: int = Field(alias="commentCount")
    comments_enabled: Optional[bool] = Field(None, alias="commentsEnabled")

    # User information
    created_by: Optional[str] = Field(None, alias="createdBy")
    updated_by: Optional[str] = Field(None, alias="updatedBy")

class QueryEvent(BaseModel):
    # Identifiers and description
    id: str
    ticker: Optional[str] = None
    slug: str
    title: str
    description: Optional[str] = None

    # Datetime
    start_date: Optional[datetime] = Field(None, alias="startDate")
    creation_date: Optional[datetime] = Field(None, alias="creationDate")
    end_date: Optional[datetime] = Field(None, alias="endDate")
    created_at: datetime = Field(alias="createdAt")
    updated_at: Optional[datetime] = Field(None, alias="updatedAt")

    # Resolution info
    resolution_source: Optional[str] = Field(None, alias="resolutionSource")

    # Media URLs
    image: Optional[str] = None
    icon: Optional[str] = None

    # Status flags
    active: bool
    closed: bool
    archived: Optional[bool] = None
    new: Optional[bool] = None
    featured: Optional[bool] = None
    restricted: Optional[bool] = None
    automatically_active: Optional[bool] = Field(None, alias="automaticallyActive")
    pending_deployment: Optional[bool] = Field(None, alias="pendingDeployment")
    deploying: Optional[bool] = None

    # Financial metrics
    liquidity: Optional[float] = None
    liquidity_clob: Optional[float] = Field(None, alias="liquidityClob")
    volume: Optional[float] = None
    open_interest: Optional[int] = Field(None, alias="openInterest")
    competitive: Optional[float] = None
    volume_24hr: Optional[float] = Field(None, alias="volume24hr")
    volume_1wk: Optional[float] = Field(None, alias="volume1wk")
    volume_1mo: Optional[float] = Field(None, alias="volume1mo")
    volume_1yr: Optional[float] = Field(None, alias="volume1yr")

    # User interaction metrics
    comment_count: Optional[int] = Field(None, alias="commentCount")

    # Related entities
    markets: Optional[list[GammaMarket]] = None
    tags: Optional[list[Tag]] = None

    # Feature toggles
    enable_order_book: Optional[bool] = Field(None, alias="enableOrderBook")
    enable_neg_risk: bool = Field(alias="enableNegRisk")
    neg_risk_augmented: Optional[bool] = Field(None, alias="negRiskAugmented")
    cyom: bool
    show_all_outcomes: bool = Field(alias="showAllOutcomes")
    show_market_images: bool = Field(alias="showMarketImages")

class Pagination(BaseModel):
    has_more: bool = Field(alias="hasMore")
    total_results: int = Field(alias="totalResults")

class EventList(BaseModel):
    events: Optional[list[QueryEvent]] = None
    pagination: Pagination

class QueryMarket(BaseModel):
    # Identifiers
    slug: str
    question: str
    group_item_title: Optional[str] = Field(None, alias="groupItemTitle")

    # Market data
    outcomes: Optional[list] = None
    outcome_prices: Optional[Json[list[float]] | list[float]] = Field(None, alias="outcomePrices")
    last_trade_price: Optional[float] = Field(None, alias="lastTradePrice")
    best_ask: Optional[float] = Field(None, alias="bestAsk")
    best_bid: Optional[float] = Field(None, alias="bestBid")
    spread: float

    # Status flags
    closed: bool
    archived: bool



# Access levels
L0 = 0
L1 = 1
L2 = 2


CREDENTIAL_CREATION_WARNING = """🚨🚨🚨
Your credentials CANNOT be recovered after they've been created.
Be sure to store them safely!
🚨🚨🚨"""


L1_AUTH_UNAVAILABLE = "A private key is needed to interact with this endpoint!"

L2_AUTH_UNAVAILABLE = "API Credentials are needed to interact with this endpoint!"

ZERO_ADDRESS = "0x0000000000000000000000000000000000000000"
HASH_ZERO = "0x0000000000000000000000000000000000000000000000000000000000000000"

AMOY = 80002
POLYGON = 137

END_CURSOR = "LTE="

BUY = "BUY"
SELL = "SELL"



from datetime import datetime
from typing import Literal, Optional, Union
from urllib.parse import urljoin

import httpx

from ..types.common import EthAddress, TimeseriesPoint
from ..types.data_types import (
    Activity,
    HolderResponse,
    Position,
    Trade,
    UserMetric,
    UserRank,
    ValueResponse,
)


class PolymarketDataClient:
    def __init__(self, base_url: str = "https://data-api.polymarket.com"):
        self.base_url = base_url
        self.client = httpx.Client(http2=True, timeout=30.0)

    def _build_url(self, endpoint: str) -> str:
        return urljoin(self.base_url, endpoint)

    def get_positions(
            self,
            user: str,
            condition_id: Optional[Union[str, list[str]]] = None,
            size_threshold: float = 1.0,
            redeemable: Optional[bool] = None,
            mergeable: Optional[bool] = None,
            title: Optional[str] = None,
            limit: int = 100,
            offset: int = 0,
            sort_by: Literal[
                "TOKENS",
                "CURRENT",
                "INITIAL",
                "CASHPNL",
                "PERCENTPNL",
                "TITLE",
                "RESOLVING",
                "PRICE",
            ] = "TOKENS",
            sort_direction: Literal["ASC", "DESC"] = "DESC",
    ) -> list[Position]:
        params = {
            "user": user,
            "sizeThreshold": size_threshold,
            "limit": min(limit, 500),
            "offset": offset,
        }
        if isinstance(condition_id, str):
            params["market"] = condition_id
        if isinstance(condition_id, list):
            params["market"] = ",".join(condition_id)
        if redeemable is not None:
            params["redeemable"] = redeemable
        if mergeable is not None:
            params["mergeable"] = mergeable
        if title:
            params["title"] = title
        if sort_by:
            params["sortBy"] = sort_by
        if sort_direction:
            params["sortDirection"] = sort_direction

        response = self.client.get(self._build_url("/positions"), params=params)
        response.raise_for_status()
        return [Position(**pos) for pos in response.json()]

    def get_trades(
            self,
            limit: int = 100,
            offset: int = 0,
            taker_only: bool = True,
            filter_type: Optional[Literal["CASH", "TOKENS"]] = None,
            filter_amount: Optional[float] = None,
            condition_id: Optional[str] = None,
            user: Optional[str] = None,
            side: Optional[Literal["BUY", "SELL"]] = None,
    ) -> list[Trade]:
        params = {
            "limit": min(limit, 500),
            "offset": offset,
            "takerOnly": taker_only,
        }
        if filter_type:
            params["filterType"] = filter_type
        if filter_amount:
            params["filterAmount"] = filter_amount
        if condition_id:
            params["market"] = condition_id
        if user:
            params["user"] = user
        if side:
            params["side"] = side

        response = self.client.get(self._build_url("/trades"), params=params)
        response.raise_for_status()
        return [Trade(**trade) for trade in response.json()]

    def get_activity(
            self,
            user: str,
            limit: int = 100,
            offset: int = 0,
            condition_id: Optional[Union[str, list[str]]] = None,
            type: Optional[Union[Literal["TRADE", "SPLIT", "MERGE", "REDEEM", "REWARD", "CONVERSION"], list[Literal["TRADE", "SPLIT", "MERGE", "REDEEM", "REWARD", "CONVERSION"]]]] = None,
            start: Optional[datetime] = None,
            end: Optional[datetime] = None,
            side: Optional[Literal["BUY", "SELL"]] = None,
            sort_by: Literal["TIMESTAMP", "TOKENS", "CASH"] = "TIMESTAMP",
            sort_direction: Literal["ASC", "DESC"] = "DESC",
    ) -> list[Activity]:
        params = {"user": user, "limit": min(limit, 500), "offset": offset}
        if isinstance(condition_id, str):
            params["market"] = condition_id
        if isinstance(condition_id, list):
            params["market"] = ",".join(condition_id)
        if isinstance(type, str):
            params["type"] = type
        if isinstance(type, list):
            params["type"] = ",".join(type)
        if start:
            params["start"] = int(start.timestamp())
        if end:
            params["end"] = int(end.timestamp())
        if side:
            params["side"] = side
        if sort_by:
            params["sortBy"] = sort_by
        if sort_direction:
            params["sortDirection"] = sort_direction

        response = self.client.get(self._build_url("/activity"), params=params)
        response.raise_for_status()
        return [Activity(**activity) for activity in response.json()]

    def get_holders(
            self,
            condition_id: str,
            limit: int = 20,
    ) -> list[HolderResponse]:
        """Takes in a condition_id and returns a list of at most 20 top holders for each corresponding token_id."""
        params = {"market": condition_id, "limit": limit}
        response = self.client.get(self._build_url("/holders"), params=params)
        response.raise_for_status()
        return [
            HolderResponse(**holder_data) for holder_data in response.json()
        ]

    def get_value(
            self,
            user: str,
            condition_ids: Optional[Union[str, list[str]]] = None,
    ) -> ValueResponse:
        """
        Get the current value of a user's position in a set of markets.

        Takes in condition_id as:
            - None      --> total value of positions
            - str       --> value of position
            - list[str] --> sum of the values of positions.
        """
        params = {"user": user}
        if isinstance(condition_ids, str):
            params["market"] = condition_ids
        if isinstance(condition_ids, list):
            params["market"] = ",".join(condition_ids)

        response = self.client.get(self._build_url("/value"), params=params)
        response.raise_for_status()
        return ValueResponse(**response.json()[0])

    # website endpoints

    def get_pnl(
            self,
            user: EthAddress,
            period: Literal["all", "1m", "1w", "1d"] = "all",
            frequency: Literal["1h", "3h", "12h", "1d"] = "1h",
    ) -> list[TimeseriesPoint]:
        """Get a user's PnL timeseries in the last day, week, month or all with a given frequency."""
        params = {
            "user_address": user,
            "interval": period,
            "fidelity": frequency,
        }

        response = self.client.get("https://user-pnl-api.polymarket.com/user-pnl", params=params)
        response.raise_for_status()
        return [TimeseriesPoint(**point) for point in response.json()]

    def get_user_metric(self, user: EthAddress, metric: Literal["profit", "volume"] = "profit", window: Literal["1d", "7d", "30d", "all"] = "all"):
        """Get a user's overall profit or volume in the last day, week, month or all."""
        params = {
            "address": user,
            "window": window,
            "limit": 1,
        }
        response = self.client.get("https://lb-api.polymarket.com/" + metric, params=params)
        response.raise_for_status()
        return UserMetric(**response.json()[0])

    def get_leaderboard_user_rank(self, user: EthAddress, metric: Literal["profit", "volume"] = "profit", window: Literal["1d", "7d", "30d", "all"] = "all"):
        """Get a user's rank on the leaderboard by profit or volume."""
        params = {
            "address": user,
            "window": window,
            "rankType": "pnl" if metric == "profit" else "vol",
        }
        response = self.client.get("https://lb-api.polymarket.com/rank", params=params)
        response.raise_for_status()
        return UserRank(**response.json()[0])

    def get_leaderboard_top_users(self, metric: Literal["profit", "volume"] = "profit", window: Literal["1d", "7d", "30d", "all"] = "all", limit: int = 100):
        """Get the leaderboard of the top at most 100 users by profit or volume."""
        params = {
            "window": window,
            "limit": limit,
        }
        response = self.client.get("https://lb-api.polymarket.com/" + metric, params=params)
        response.raise_for_status()
        return [UserMetric(**user) for user in response.json()]


    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.client.close()



from eth_account import Account


class Signer:
    def __init__(self, private_key: str, chain_id: int):
        if private_key is None:
            msg = "private_key must not be None"
            raise ValueError(msg)
        if chain_id is None:
            msg = "chain_id must not be None"
            raise ValueError(msg)

        self.private_key = private_key
        self.account = Account.from_key(private_key)
        self.chain_id = chain_id

    def address(self):
        return self.account.address

    def get_chain_id(self):
        return self.chain_id

    def sign(self, message_hash):
        """Signs a message hash."""
        return Account.unsafe_sign_hash(message_hash, self.private_key).signature.hex()



import re
from datetime import datetime
from typing import Annotated

from pydantic import AfterValidator, BaseModel, BeforeValidator, Field


def validate_keccak256(v: str) -> str:
    if not re.match(r"^0x[a-fA-F0-9]{64}$", v):
        msg = "Invalid Keccak256 hash format"
        raise ValueError(msg)
    return v

def parse_timestamp(v: str) -> datetime:
    if isinstance(v, datetime):
        return v
    # Normalize '+00' to '+0000' for timezone
    if v.endswith("+00"):
        v = v[:-3] + "+0000"

    # Pad fractional seconds to 6 digits if present
    if "." in v:
        dot_pos = v.find(".")
        tz_pos = v.find("+", dot_pos)  # Find timezone start after '.'
        if tz_pos == -1:
            tz_pos = v.find("-", dot_pos)

        if tz_pos != -1:
            frac = v[dot_pos+1:tz_pos]
            if len(frac) < 6:
                frac = frac.ljust(6, "0")
                v = f"{v[:dot_pos+1]}{frac}{v[tz_pos:]}"

    # Try parsing with and without microseconds
    for fmt in ("%Y-%m-%d %H:%M:%S.%f%z", "%Y-%m-%d %H:%M:%S%z"):
        try:
            return datetime.strptime(v, fmt) # noqa: DTZ007
        except ValueError:
            continue
    msg = f"Time data '{v}' does not match expected formats."
    raise ValueError(msg)


TimestampWithTZ = Annotated[datetime, BeforeValidator(parse_timestamp)]
EthAddress = Annotated[str, Field(pattern=r"^0x[A-Fa-f0-9]{40}$")]
Keccak256 = Annotated[str, AfterValidator(validate_keccak256)]
EmptyString = Annotated[str, Field(pattern=r"^$", description="An empty string")]

class TimeseriesPoint(BaseModel):
    value: float
    timestamp: datetime



CUSTOM_ERROR_DICT = {
    "0x3a81d6fc": "AlreadyRegistered()",
    "0xcd4e6167": "FeeTooHigh()",
    "0x66f8620a": "InvalidComplement()",
    "0x756688fe": "InvalidNonce()",
    "0x8baa579f": "InvalidSignature()",
    "0x3f6cc768": "InvalidTokenId()",
    "0xe2cc6ad6": "MakingGtRemaining()",
    "0xa0b94465": "MismatchedTokenIds()",
    "0x7bfa4b9f": "NotAdmin()",
    "0x7f9a6f46": "NotCrossing()",
    "0x7c214f04": "NotOperator()",
    "0x30cd7471": "NotOwner()",
    "0x5211a079": "NotTaker()",
    "0xc56873ba": "OrderExpired()",
    "0x7b38b76e": "OrderFilledOrCancelled()",
    "0x9e87fac8": "Paused()",
    "0xdf4d8080": "TooLittleTokensReceived()",
    "0xb36eb036": "DeterminedFlagAlreadySet()",
    "0xfc520af5": "FeeBipsOutOfBounds()",
    "0x4e23d035": "IndexOutOfBounds()",
    "0x9667d381": "InvalidIndexSet()",
    "0xff633a38": "LengthMismatch()",
    "0xe3b02389": "MarketAlreadyDetermined()",
    "0xd2b4c0c6": "MarketAlreadyPrepared()",
    "0xb6649495": "MarketNotPrepared()",
    "0x24f2dfe7": "NoConvertiblePositions()",
    "0xcd7769ff": "NotApprovedForAll()",
    "0x80fee105": "OnlyOracle()",
    "0x155a4b5c": "UnexpectedCollateralToken()",
    }
